# Copyright (C) 2015-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Handle interprocess communication using 0MQ
"""


__author__ = "Damon Lynch"
__copyright__ = "Copyright 2015-2024, Damon Lynch"

import argparse
import logging
import os
import pickle
import shlex
import sys
import time
from collections import deque
from collections.abc import Sequence
from typing import Any

import psutil
import zmq
import zmq.log.handlers
from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QImage, QPixmap

if zmq.pyzmq_version_info()[0] < 17:
    from zmq.eventloop import ioloop
else:
    try:
        from tornado import ioloop
    except ImportError:
        from zmq.eventloop import ioloop  # note: deprecated in pyzmq 17.0.0

from zmq.eventloop.zmqstream import ZMQStream

from raphodo.constants import (
    BackupStatus,
    CameraErrorCode,
    ExtractionProcessing,
    ExtractionTask,
    RenameAndMoveStatus,
)
from raphodo.devices import Device
from raphodo.folderspreview import FoldersPreview
from raphodo.iplogging import ZeroMQSocketHandler
from raphodo.problemnotification import (
    BackingUpProblems,
    CopyingProblems,
    RenamingProblems,
    ScanProblems,
)
from raphodo.proximity import TemporalProximityGroups
from raphodo.rpdfile import FileSizeSum, FileTypeCounter, Photo, RPDFile, Video
from raphodo.storage.storage import StorageSpace
from raphodo.ui.viewutils import ThumbnailDataForProximity
from raphodo.utilities import CacheDirs, set_pdeathsig

logger = logging.getLogger()


def make_filter_from_worker_id(worker_id: int| str) -> bytes:
    r"""
    Returns a python byte string from an integer or string

    >>> make_filter_from_worker_id(54)
    b'54'

    >>> make_filter_from_worker_id('54')
    b'54'
    """
    if isinstance(worker_id, int):
        return str(worker_id).encode()
    if isinstance(worker_id, str):
        return worker_id.encode()
    raise (TypeError)


def create_identity(worker_type: str, identity: str) -> bytes:
    r"""Generate identity for a worker's 0mq socket.

    >>> create_identity('Worker', '1')
    b'Worker-1'
    >>> create_identity('Thumbnail Extractor', '2')
    b'Thumbnail-Extractor-2'
    >>> create_identity('Thumbnail Extractor Plus', '22 2')
    b'Thumbnail-Extractor-Plus-22-2'
    """

    # Replace any whitespace in the strings with a hyphen
    return "{}-{}".format(
        "-".join(worker_type.split()), "-".join(identity.split())
    ).encode()


def get_worker_id_from_identity(identity: bytes) -> int:
    r"""Extract worker id from the identity used in a 0mq socket

    >>> get_worker_id_from_identity(b'Worker-1')
    1
    >>> get_worker_id_from_identity(b'Thumbnail-Extractor-2')
    2
    >>> get_worker_id_from_identity(b'Thumbnail-Extractor-Plus-22-2')
    2
    """
    return int(identity.decode().split("-")[-1])


def create_inproc_msg(
    cmd: bytes, worker_id: int|None = None, data: Any | None = None
) -> list[bytes]:
    """
    Create a list of three values to be sent via a PAIR socket
    between main and child threads using 0MQ.
    """

    worker_id = make_filter_from_worker_id(worker_id) if worker_id is not None else b""

    data = b"" if data is None else pickle.dumps(data, pickle.HIGHEST_PROTOCOL)

    return [cmd, worker_id, data]


class ThreadNames:
    rename = "rename"
    scan = "scan"
    copy = "copy"
    backup = "backup"
    thumbnail_daemon = "thumbnail_daemon"
    thumbnailer = "thumbnailer"
    offload = "offload"
    logger = "logger"
    load_balancer = "load_balancer"
    new_version = "new_version"


class ProcessManager:
    def __init__(self, logging_port: int, thread_name: str) -> None:

        super().__init__()

        self.logging_port = logging_port

        self.processes = {}  # type: dict[int, psutil.Process]
        self._process_to_run = ""  # Implement in subclass

        self.thread_name = thread_name

        # Monitor which workers we have running
        self.workers = []  # type: list[int]

    def _get_cmd(self) -> str:
        return "{} {}".format(
            sys.executable,
            os.path.join(
                os.path.abspath(os.path.dirname(__file__)), self._process_to_run
            ),
        )

    def _get_command_line(self, worker_id: int) -> str:
        """
        Implement in subclass
        """
        return ""

    def add_worker(self, worker_id: int) -> None:

        command_line = self._get_command_line(worker_id)
        args = shlex.split(command_line)

        # run command immediately, without waiting a reply, and instruct the Linux
        # kernel to send a terminate signal should this process unexpectedly die
        try:
            proc = psutil.Popen(args, preexec_fn=set_pdeathsig())
        except OSError as e:
            logging.critical("Failed to start process: %s", command_line)
            logging.critical("OSError [Errno %s]: %s", e.errno, e.strerror)
            if e.errno == 8:
                logging.critical(
                    "Script shebang line might be malformed or missing: %s",
                    self._get_cmd(),
                )
            sys.exit(1)
        logging.debug("Started '%s' with pid %s", command_line, proc.pid)

        # Add to list of running workers
        self.workers.append(worker_id)
        self.processes[worker_id] = proc

    def forcefully_terminate(self) -> None:
        """
        Forcefully terminate any running child processes.
        """

        zombie_processes = [
            p
            for p in self.processes.values()
            if p.is_running() and p.status() == psutil.STATUS_ZOMBIE
        ]
        running_processes = [
            p
            for p in self.processes.values()
            if p.is_running() and p.status() != psutil.STATUS_ZOMBIE
        ]
        if hasattr(self, "_process_name"):
            logging.debug(
                "Forcefully terminating processes for %s: %s zombies, %s running.",
                self._process_name,
                len(zombie_processes),
                len(running_processes),
            )

        for p in zombie_processes:  # type: psutil.Process
            try:
                logging.debug("Killing zombie process %s with pid %s", p.name(), p.pid)
                p.kill()
            except Exception:
                logging.error("Failed to kill process with pid %s", p.pid)
        for p in running_processes:  # type: psutil.Process
            try:
                logging.debug("Terminating process %s with pid %s", p.name(), p.pid)
                p.terminate()
            except Exception:
                logging.error("Terminating process with pid %s failed", p.pid)
        gone, alive = psutil.wait_procs(running_processes, timeout=2)
        for p in alive:
            try:
                logging.debug("Killing zombie process %s with pid %s", p.name(), p.pid)
                p.kill()
            except Exception:
                logging.error("Failed to kill process with pid %s", p.pid)

    def process_alive(self, worker_id: int) -> bool:
        """
        Process IDs are reused by the system. Check to make sure
        a new process has not been created with the same process id.

        :param worker_id: the process to check
        :return True if the process is the same, False otherwise
        """

        return self.processes[worker_id].is_running()


class PullPipelineManager(ProcessManager, QObject):
    """
    Base class from which more specialized 0MQ classes are derived.

    Receives data into its sink via a ZMQ PULL socket, but does not
    specify how workers should be sent data.

    Outputs signals using Qt.
    """

    message = pyqtSignal(str)  # Derived class will change this
    sinkStarted = pyqtSignal()
    workerFinished = pyqtSignal(int)
    workerStopped = pyqtSignal(int)
    receiverPortSignal = pyqtSignal(int)

    def __init__(self, logging_port: int, thread_name: str) -> None:
        super().__init__(logging_port=logging_port, thread_name=thread_name)

    def _start_sockets(self) -> None:

        context = zmq.Context.instance()

        # Subclasses must define the type of port they need to send messages
        self.ventilator_socket = None
        self.ventilator_port = None

        # Sink socket to receive results of the workers
        self.receiver_socket = context.socket(zmq.PULL)
        self.receiver_port = self.receiver_socket.bind_to_random_port("tcp://*")

        # Socket to communicate directly with the sink, bypassing the workers
        self.terminate_socket = context.socket(zmq.PUSH)
        self.terminate_socket.connect(f"tcp://localhost:{self.receiver_port}")

        # Socket to receive commands from the main thread
        self.thread_controller = context.socket(zmq.PAIR)
        self.thread_controller.connect(f"inproc://{self.thread_name}")

        self.terminating = False

    @pyqtSlot()
    def run_sink(self) -> None:
        logging.debug("Running sink for %s", self._process_name)

        self._start_sockets()

        poller = zmq.Poller()
        poller.register(self.receiver_socket, zmq.POLLIN)
        poller.register(self.thread_controller, zmq.POLLIN)

        self.receiverPortSignal.emit(self.receiver_port)
        self.sinkStarted.emit()

        while True:
            try:
                socks = dict(poller.poll())
            except KeyboardInterrupt:
                break
            if self.receiver_socket in socks:
                # Receive messages from the workers
                # (or the terminate socket)
                worker_id, directive, content = self.receiver_socket.recv_multipart()

                if directive == b"cmd":
                    command = content
                    assert command in (b"STOPPED", b"FINISHED", b"KILL")
                    if command == b"KILL":
                        # Terminate immediately, without regard for any
                        # incoming messages. This message is only sent
                        # from this manager to itself, using the
                        # self.terminate_socket
                        logging.debug(f"{self._process_name} is terminating")
                        break
                    # This worker is done; remove from monitored workers and
                    # continue
                    worker_id = int(worker_id)
                    if command == b"STOPPED":
                        logging.debug(
                            "%s worker %s has stopped", self._process_name, worker_id
                        )
                        self.workerStopped.emit(worker_id)
                    else:
                        # Worker has finished its work
                        self.workerFinished.emit(worker_id)
                    self.workers.remove(worker_id)
                    del self.processes[worker_id]
                    if not self.workers:
                        logging.debug(
                            f"{self._process_name} currently has no workers"
                        )
                    if not self.workers and self.terminating:
                        logging.debug(f"{self._process_name} is exiting")
                        break
                else:
                    assert directive == b"data"
                    self.content = content
                    self.process_sink_data()

            if self.thread_controller in socks:
                # Receive messages from the main Rapid Photo Downloader thread
                self.process_thread_directive()

    def process_thread_directive(self) -> None:
        directive, worker_id, data = self.thread_controller.recv_multipart()

        # Directives: START, STOP, TERMINATE, SEND_TO_WORKER, STOP_WORKER, START_WORKER
        if directive == b"START":
            self.start()
        elif directive == b"START_WORKER":
            self.start_worker(worker_id=worker_id, data=data)
        elif directive == b"SEND_TO_WORKER":
            self.send_message_to_worker(worker_id=worker_id, data=data)
        elif directive == b"STOP":
            self.stop()
        elif directive == b"STOP_WORKER":
            self.stop_worker(worker_id=worker_id)
        elif directive == b"PAUSE":
            self.pause()
        elif directive == b"RESUME":
            self.resume(worker_id=worker_id)
        elif directive == b"TERMINATE":
            self.forcefully_terminate()
        else:
            logging.critical("%s received unknown directive %s", directive.decode())

    def process_sink_data(self) -> None:
        data = pickle.loads(self.content)
        self.message.emit(data)

    def terminate_sink(self) -> None:
        self.terminate_socket.send_multipart([b"0", b"cmd", b"KILL"])

    def _get_ventilator_start_message(self, worker_id: bytes) -> list:
        return [worker_id, b"cmd", b"START"]

    def start(self) -> None:
        logging.critical(
            "Member function start() not implemented in child class of %s",
            self._process_name,
        )

    def start_worker(self, worker_id: bytes, data: bytes) -> None:
        logging.critical(
            "Member function start_worker() not implemented in child class of %s",
            self._process_name,
        )

    def stop(self) -> None:
        logging.critical(
            "Member function stop() not implemented in child class of %s",
            self._process_name,
        )

    def stop_worker(self, worker_id: int) -> None:
        logging.critical(
            "Member function stop_worker() not implemented in child class of %s",
            self._process_name,
        )

    def pause(self) -> None:
        logging.critical(
            "Member function pause() not implemented in child class of %s",
            self._process_name,
        )

    def resume(self, worker_id: bytes|None) -> None:
        logging.critical(
            "Member function stop_worker() not implemented in child class of %s",
            self._process_name,
        )

    def send_message_to_worker(
        self, data: bytes, worker_id: bytes|None = None
    ) -> None:
        if self.terminating:
            logging.debug(
                "%s not sending message to worker because manager is terminated",
                self._process_name,
            )
            return
        if not self.workers:
            logging.debug(
                "%s not sending message to worker because there are no workers",
                self._process_name,
            )
            return

        assert isinstance(data, bytes)

        message = [worker_id, b"data", data] if worker_id else [b"data", data]
        self.ventilator_socket.send_multipart(message)

    def forcefully_terminate(self) -> None:
        """
        Forcefully terminate any child processes and clean up.

        Shuts down the sink too.
        """

        super().forcefully_terminate()
        self.terminate_sink()


class LoadBalancerWorkerManager(ProcessManager):
    def __init__(
        self, no_workers: int, backend_port: int, sink_port: int, logging_port: int
    ) -> None:
        super().__init__(logging_port=logging_port, thread_name="")
        self.no_workers = no_workers
        self.backend_port = backend_port
        self.sink_port = sink_port

    def _get_command_line(self, worker_id: int) -> str:
        cmd = self._get_cmd()

        return "{} --request {} --send {} --identity {} --logging {}".format(
            cmd, self.backend_port, self.sink_port, worker_id, self.logging_port
        )

    def start_workers(self) -> None:
        for worker_id in range(self.no_workers):
            self.add_worker(worker_id)

    def zombie_workers(self) -> list[int]:
        return [
            worker_id
            for worker_id in self.workers
            if self.processes[worker_id].status() == psutil.STATUS_ZOMBIE
        ]


class LRUQueue:
    """LRUQueue class using ZMQStream/IOLoop for event dispatching"""

    def __init__(
        self,
        backend_socket: zmq.Socket,
        frontend_socket: zmq.Socket,
        controller_socket: zmq.Socket,
        worker_type: str,
        process_manager: LoadBalancerWorkerManager,
    ) -> None:

        self.worker_type = worker_type
        self.process_manager = process_manager
        self.workers = deque()
        self.terminating = False
        self.terminating_workers = set()  # type: set[bytes]
        self.stopped_workers = set()  # type: set[int]

        self.backend = ZMQStream(backend_socket)
        self.frontend = ZMQStream(frontend_socket)
        self.controller = ZMQStream(controller_socket)
        self.backend.on_recv(self.handle_backend)
        self.controller.on_recv(self.handle_controller)

        self.loop = ioloop.IOLoop.instance()

    def handle_controller(self, msg):
        self.terminating = True

        while len(self.workers):
            worker_identity = self.workers.popleft()

            logging.debug(
                "%s load balancer sending stop cmd to worker %s",
                self.worker_type,
                worker_identity.decode(),
            )
            self.backend.send_multipart([worker_identity, b"", b"cmd", b"STOP"])
            self.terminating_workers.add(worker_identity)

        self.loop.add_timeout(time.time() + 3, self.loop.stop)

    def handle_backend(self, msg):
        # Queue worker address for LRU routing
        worker_identity, empty, client_addr = msg[:3]

        # add worker back to the list of workers
        self.workers.append(worker_identity)

        zw = self.process_manager.zombie_workers()
        if zw:
            logging.critical("%s dead thumbnail extractors", len(zw))

        # Second frame is empty
        assert empty == b""

        if msg[-1] == b"STOPPED" and self.terminating:
            worker_id = get_worker_id_from_identity(worker_identity)
            self.stopped_workers.add(worker_id)
            self.terminating_workers.remove(worker_identity)
            if len(self.terminating_workers) == 0:
                for worker_id in self.stopped_workers:
                    p = self.process_manager.processes[
                        worker_id
                    ]  # type: psutil.Process
                    if p.is_running():
                        pid = p.pid
                        if p.status() != psutil.STATUS_SLEEPING:
                            logging.debug(
                                "Waiting on %s process %s...", p.status(), pid
                            )
                            os.waitpid(pid, 0)
                            logging.debug("...process %s is finished", pid)
                        else:
                            logging.debug("Process %s is sleeping", pid)
                self.loop.add_timeout(time.time() + 0.5, self.loop.stop)

        if len(self.workers) == 1:
            # on first recv, start accepting frontend messages
            self.frontend.on_recv(self.handle_frontend)

    def handle_frontend(self, request):
        #  Dequeue and drop the next worker address
        worker_identity = self.workers.popleft()

        message = [worker_identity, b""] + request
        self.backend.send_multipart(message)
        if len(self.workers) == 0:
            # stop receiving until workers become available again
            self.frontend.stop_on_recv()


class LoadBalancer:
    def __init__(self, worker_type: str, process_manager) -> None:

        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("--receive", required=True)
        self.parser.add_argument("--send", required=True)
        self.parser.add_argument("--controller", required=True)
        self.parser.add_argument("--logging", required=True)

        args = self.parser.parse_args()
        self.controller_port = args.controller

        context = zmq.Context()
        frontend = context.socket(zmq.PULL)
        frontend_port = frontend.bind_to_random_port("tcp://*")

        backend = context.socket(zmq.ROUTER)
        backend_port = backend.bind_to_random_port("tcp://*")

        reply = context.socket(zmq.REP)
        reply.connect(f"tcp://localhost:{args.receive}")

        controller = context.socket(zmq.PULL)
        controller.connect(f"tcp://localhost:{self.controller_port}")

        sink_port = args.send
        logging_port = args.logging

        self.logger_publisher = ProcessLoggerPublisher(
            context=context, name=worker_type, notification_port=args.logging
        )

        logging.debug(
            f"{worker_type} load balancer waiting to be notified how many workers to "
            "initialize..."
        )
        no_workers = int(reply.recv())
        logging.debug(
            f"...{worker_type} load balancer will use {no_workers} workers"
        )
        reply.send(str(frontend_port).encode())

        process_manager = process_manager(
            no_workers, backend_port, sink_port, logging_port
        )
        process_manager.start_workers()

        # create queue with the sockets
        queue = LRUQueue(backend, frontend, controller, worker_type, process_manager)  # noqa: F841

        # start reactor, which is an infinite loop
        ioloop.IOLoop.instance().start()

        # Finished infinite loop: do some housekeeping
        logging.debug("Forcefully terminating load balancer child processes")
        process_manager.forcefully_terminate()

        frontend.close()
        backend.close()


class LoadBalancerManager(ProcessManager, QObject):
    """
    Launches and requests termination of the Load Balancer process
    """

    load_balancer_started = pyqtSignal(int)

    def __init__(
        self,
        context: zmq.Context,
        no_workers: int,
        sink_port: int,
        logging_port: int,
        thread_name: str,
    ) -> None:
        super().__init__(logging_port=logging_port, thread_name=thread_name)
        self.no_workers = no_workers
        self.sink_port = sink_port
        self.context = context

    @pyqtSlot()
    def start_load_balancer(self) -> None:

        self.controller_socket = self.context.socket(zmq.PUSH)
        self.controller_port = self.controller_socket.bind_to_random_port("tcp://*")

        self.requester = self.context.socket(zmq.REQ)
        self.requester_port = self.requester.bind_to_random_port("tcp://*")

        self.thread_controller = self.context.socket(zmq.PAIR)
        self.thread_controller.connect(f"inproc://{self.thread_name}")

        worker_id = 0
        self.add_worker(worker_id)
        self.requester.send(str(self.no_workers).encode())
        self.frontend_port = int(self.requester.recv())
        self.load_balancer_started.emit(self.frontend_port)

        # wait for stop signal
        directive, worker_id, data = self.thread_controller.recv_multipart()
        assert directive == b"STOP"
        self.stop()

    def stop(self):
        self.controller_socket.send(b"STOP")

    def _get_command_line(self, worker_id: int) -> str:
        cmd = self._get_cmd()

        return "{} --receive {} --send {} --controller {} --logging {}".format(
            cmd,
            self.requester_port,
            self.sink_port,
            self.controller_port,
            self.logging_port,
        )


DAEMON_WORKER_ID = 0


class PushPullDaemonManager(PullPipelineManager):
    """
    Manage a single instance daemon worker process that waits to work on data
    issued by this manager. The data to be worked on is issued in sequence,
    one after the other.

    Because this is a single daemon process, a Push-Pull model is most
    suitable for sending the data.
    """

    def _start_sockets(self) -> None:

        super()._start_sockets()

        context = zmq.Context.instance()

        # Ventilator socket to send message to worker
        self.ventilator_socket = context.socket(zmq.PUSH)
        self.ventilator_port = self.ventilator_socket.bind_to_random_port("tcp://*")

    def stop(self) -> None:
        """
        Permanently stop the daemon process and terminate
        """

        logging.debug(f"{self._process_name} halting")
        self.terminating = True

        # Only send stop command if the process is still running
        if self.process_alive(DAEMON_WORKER_ID):
            try:
                self.ventilator_socket.send_multipart([b"cmd", b"STOP"], zmq.DONTWAIT)
            except zmq.Again:
                logging.debug(
                    "Terminating %s sink because child process did not receive message",
                    self._process_name,
                )
                self.terminate_sink()
        else:
            # The process may have crashed. Stop the sink.
            self.terminate_sink()

    def _get_command_line(self, worker_id: int) -> str:
        cmd = self._get_cmd()

        return "{} --receive {} --send {} --logging {}".format(
            cmd, self.ventilator_port, self.receiver_port, self.logging_port
        )

    def _get_ventilator_start_message(self, worker_id: int) -> list[bytes]:
        return [b"cmd", b"START"]

    def start(self) -> None:
        logging.debug("Starting worker for %s", self._process_name)
        self.add_worker(worker_id=DAEMON_WORKER_ID)


class PublishPullPipelineManager(PullPipelineManager):
    """
    Manage a collection of worker processes that wait to work on data
    issued by this manager. The data to be worked on is issued in sequence,
    one after the other, either once, or many times.

    Because there are multiple worker process, a Publish-Subscribe model is
    most suitable for sending data to workers.
    """

    def _start_sockets(self) -> None:

        super()._start_sockets()

        context = zmq.Context.instance()

        # Ventilator socket to send messages to workers on
        self.ventilator_socket = context.socket(zmq.PUB)
        self.ventilator_port = self.ventilator_socket.bind_to_random_port("tcp://*")

        # Socket to synchronize the start of each worker
        self.sync_service_socket = context.socket(zmq.REP)
        self.sync_service_port = self.sync_service_socket.bind_to_random_port("tcp://*")

        # Socket for worker control: pause, resume, stop
        self.controller_socket = context.socket(zmq.PUB)
        self.controller_port = self.controller_socket.bind_to_random_port("tcp://*")

    def stop(self) -> None:
        """
        Permanently stop all the workers and terminate
        """

        logging.debug(f"{self._process_name} halting")
        self.terminating = True
        if self.workers:
            # Signal workers they must immediately stop
            termination_signal_sent = False
            alive_workers = [
                worker_id for worker_id in self.workers if self.process_alive(worker_id)
            ]
            for worker_id in alive_workers:

                message = [make_filter_from_worker_id(worker_id), b"STOP"]
                self.controller_socket.send_multipart(message)

                message = [make_filter_from_worker_id(worker_id), b"cmd", b"STOP"]
                self.ventilator_socket.send_multipart(message)
                termination_signal_sent = True

            if not termination_signal_sent:
                self.terminate_sink()
        else:
            self.terminate_sink()

    def stop_worker(self, worker_id: bytes) -> None:
        """
        Permanently stop one worker
        """

        if int(worker_id) in self.workers:
            message = [worker_id, b"STOP"]
            self.controller_socket.send_multipart(message)
            message = [worker_id, b"cmd", b"STOP"]
            self.ventilator_socket.send_multipart(message)

    def start_worker(self, worker_id: bytes, data: bytes) -> None:

        self.add_worker(int(worker_id))

        # Send START commands until scan worker indicates it is ready to
        # receive data
        # Worker ID must be in bytes format
        while True:
            self.ventilator_socket.send_multipart(
                self._get_ventilator_start_message(worker_id)
            )
            try:
                # look for synchronization request
                self.sync_service_socket.recv(zmq.DONTWAIT)
                # send synchronization reply
                self.sync_service_socket.send(b"")
                break
            except zmq.Again:
                # Briefly pause sending out START messages
                # There is no point flooding the network
                time.sleep(0.01)

        # Send data to process to tell it what to work on
        self.send_message_to_worker(data=data, worker_id=worker_id)

    def _get_command_line(self, worker_id: int) -> str:
        cmd = self._get_cmd()

        return (
            "{} --receive {} --send {} --controller {} --syncclient {} "
            "--filter {} --logging "
            "{}".format(
                cmd,
                self.ventilator_port,
                self.receiver_port,
                self.controller_port,
                self.sync_service_port,
                worker_id,
                self.logging_port,
            )
        )

    def __len__(self) -> int:
        return len(self.workers)

    def __contains__(self, item) -> bool:
        return item in self.workers

    def pause(self) -> None:
        for worker_id in self.workers:
            message = [make_filter_from_worker_id(worker_id), b"PAUSE"]
            self.controller_socket.send_multipart(message)

    def resume(self, worker_id: bytes) -> None:
        workers = [int(worker_id)] if worker_id else self.workers
        for worker_id in workers:
            message = [make_filter_from_worker_id(worker_id), b"RESUME"]
            self.controller_socket.send_multipart(message)


class ProcessLoggerPublisher:
    """
    Setup the sockets for worker processes to send log messages to the
    main process.

    Two tasks: set up the PUB socket, and then tell the main process
    what port we're using via a second socket, and when we're closing it.
    """

    def __init__(self, context: zmq.Context, name: str, notification_port: int) -> None:

        self.logger_pub = context.socket(zmq.PUB)
        self.logger_pub_port = self.logger_pub.bind_to_random_port("tcp://*")
        self.handler = ZeroMQSocketHandler(self.logger_pub)
        self.handler.setLevel(logging.DEBUG)

        self.logger = logging.getLogger()
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(self.handler)

        self.logger_socket = context.socket(zmq.PUSH)
        self.logger_socket.connect(f"tcp://localhost:{notification_port}")
        self.logger_socket.send_multipart(
            [b"CONNECT", str(self.logger_pub_port).encode()]
        )

    def close(self):
        self.logger.removeHandler(self.handler)
        self.logger_socket.send_multipart(
            [b"DISCONNECT", str(self.logger_pub_port).encode()]
        )
        self.logger_pub.close()
        self.logger_socket.close()


class WorkerProcess:
    def __init__(self, worker_type: str) -> None:
        super().__init__()
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("--receive", required=True)
        self.parser.add_argument("--send", required=True)
        self.parser.add_argument("--logging", required=True)

    def cleanup_pre_stop(self) -> None:
        """
        Operations to run if process is stopped.

        Implement in child class if needed.
        """

        pass

    def setup_logging_pub(self, notification_port: int, name: str) -> None:
        """
        Sets up the 0MQ socket that sends out logging messages

        :param notification_port: port that should be notified about
         the new logging publisher
        :param name: descriptive name to place in the log messages
        """

        if self.worker_id is not None:
            name = f"{name}-{self.worker_id.decode()}"
        self.logger_publisher = ProcessLoggerPublisher(
            context=self.context, name=name, notification_port=notification_port
        )

    def send_message_to_sink(self) -> None:

        self.sender.send_multipart([self.worker_id, b"data", self.content])

    def initialise_process(self) -> None:
        # Wait to receive "START" message
        worker_id, directive, content = self.receiver.recv_multipart()
        assert directive == b"cmd"
        assert content == b"START"

        # send a synchronization request
        self.sync_client.send(b"")

        # wait for synchronization reply
        self.sync_client.recv()

        # Receive next "START" message and discard, looking for data message
        while True:
            worker_id, directive, content = self.receiver.recv_multipart()
            if directive == b"data":
                break
            else:
                assert directive == b"cmd"
                assert content == b"START"

        self.content = content

    def do_work(self):
        pass


class DaemonProcess(WorkerProcess):
    """
    Single instance
    """

    def __init__(self, worker_type: str) -> None:
        super().__init__(worker_type)

        args = self.parser.parse_args()

        self.context = zmq.Context()
        # Socket to send messages along the pipe to
        self.sender = self.context.socket(zmq.PUSH)
        self.sender.set_hwm(10)
        self.sender.connect(f"tcp://localhost:{args.send}")

        self.receiver = self.context.socket(zmq.PULL)
        self.receiver.connect(f"tcp://localhost:{args.receive}")

        self.worker_id = None

        self.setup_logging_pub(notification_port=args.logging, name=worker_type)

    def run(self) -> None:
        pass

    def check_for_command(self, directive: bytes, content: bytes) -> None:
        if directive == b"cmd":
            assert content == b"STOP"
            self.cleanup_pre_stop()
            # signal to sink that we've terminated before finishing
            self.sender.send_multipart(
                [make_filter_from_worker_id(DAEMON_WORKER_ID), b"cmd", b"STOPPED"]
            )
            sys.exit(0)

    def send_message_to_sink(self) -> None:
        # Must use a dummy value for the worker id, as there is only ever one
        # instance.
        self.sender.send_multipart(
            [make_filter_from_worker_id(DAEMON_WORKER_ID), b"data", self.content]
        )


class WorkerInPublishPullPipeline(WorkerProcess):
    """
    Worker counterpart to PublishPullPipelineManager; multiple instance.
    """

    def __init__(self, worker_type: str) -> None:
        super().__init__(worker_type)
        self.add_args()

        args = self.parser.parse_args()

        subscription_filter = self.worker_id = args.filter.encode()
        self.context = zmq.Context()

        self.setup_sockets(args, subscription_filter)
        self.setup_logging_pub(notification_port=args.logging, name=worker_type)

        self.initialise_process()
        self.do_work()

    def add_args(self) -> None:
        self.parser.add_argument("--filter", required=True)
        self.parser.add_argument("--syncclient", required=True)
        self.parser.add_argument("--controller", required=True)

    def setup_sockets(self, args, subscription_filter: bytes) -> None:

        # Socket to send messages along the pipe to
        self.sender = self.context.socket(zmq.PUSH)
        self.sender.set_hwm(10)
        self.sender.connect(f"tcp://localhost:{args.send}")

        # Socket to receive messages from the pipe
        self.receiver = self.context.socket(zmq.SUB)
        self.receiver.connect(f"tcp://localhost:{args.receive}")
        self.receiver.setsockopt(zmq.SUBSCRIBE, subscription_filter)

        # Socket to receive controller messages: stop, pause, resume
        self.controller = self.context.socket(zmq.SUB)
        self.controller.connect(f"tcp://localhost:{args.controller}")
        self.controller.setsockopt(zmq.SUBSCRIBE, subscription_filter)

        # Socket to synchronize the start of receiving data from upstream
        self.sync_client = self.context.socket(zmq.REQ)
        self.sync_client.connect(f"tcp://localhost:{args.syncclient}")

    def check_for_command(self, directive: bytes, content) -> None:
        if directive == b"cmd":
            try:
                assert content == b"STOP"
            except AssertionError:
                logging.critical(
                    "Expected STOP command but instead got %s", content.decode()
                )
            else:
                self.cleanup_pre_stop()
                self.disconnect_logging()
                # signal to sink that we've terminated before finishing
                self.sender.send_multipart([self.worker_id, b"cmd", b"STOPPED"])
                sys.exit(0)

    def check_for_controller_directive(self) -> None:
        try:
            # Don't block if the process is running regularly
            # If there is no command, an exception will occur
            worker_id, command = self.controller.recv_multipart(zmq.DONTWAIT)
            assert command in [b"PAUSE", b"STOP"]
            assert worker_id == self.worker_id

            if command == b"PAUSE":
                # Because the process is paused, do a blocking read to
                # wait for the next command
                worker_id, command = self.controller.recv_multipart()
                assert command in [b"RESUME", b"STOP"]
            if command == b"STOP":
                self.cleanup_pre_stop()
                # before finishing, signal to sink that we've terminated
                self.sender.send_multipart([self.worker_id, b"cmd", b"STOPPED"])
                sys.exit(0)
        except zmq.Again:
            pass  # Continue working

    def resume_work(self) -> None:
        worker_id, command = self.controller.recv_multipart()
        assert command in [b"RESUME", b"STOP"]
        if command == b"STOP":
            self.cleanup_pre_stop()
            self.disconnect_logging()
            # before finishing, signal to sink that we've terminated
            self.sender.send_multipart([self.worker_id, b"cmd", b"STOPPED"])
            sys.exit(0)

    def disconnect_logging(self) -> None:
        self.logger_publisher.close()

    def send_finished_command(self) -> None:
        self.sender.send_multipart([self.worker_id, b"cmd", b"FINISHED"])


class LoadBalancerWorker:
    def __init__(self, worker_type: str) -> None:
        super().__init__()
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("--request", required=True)
        self.parser.add_argument("--send", required=True)
        self.parser.add_argument("--identity", required=True)
        self.parser.add_argument("--logging", required=True)

        args = self.parser.parse_args()

        self.context = zmq.Context()

        self.requester = self.context.socket(zmq.REQ)
        self.identity = create_identity(worker_type, args.identity)
        self.requester.identity = self.identity
        self.requester.connect(f"tcp://localhost:{args.request}")

        # Sender is located in the main process. It is where output (messages)
        # from this process are sent to.
        self.sender = self.context.socket(zmq.PUSH)
        self.sender.connect(f"tcp://localhost:{args.send}")

        self.logger_publisher = ProcessLoggerPublisher(
            context=self.context, name=worker_type, notification_port=args.logging
        )

        # Tell the load balancer we are ready for work
        self.requester.send(b"READY")
        self.do_work()

    def do_work(self) -> None:
        # Implement in subclass
        pass

    def cleanup_pre_stop(self) -> None:
        """
        Operations to run if the process is stopped.

        Implement in child class if needed.
        """

        pass

    def exit(self):
        self.cleanup_pre_stop()
        identity = self.requester.identity.decode()
        # signal to load balancer that we've terminated before finishing
        self.requester.send_multipart([b"", b"", b"STOPPED"])
        self.requester.close()
        self.sender.close()
        self.logger_publisher.close()
        self.context.term()
        logging.debug("%s with pid %s stopped", identity, os.getpid())
        sys.exit(0)

    def check_for_command(self, directive: bytes, content: bytes):
        if directive == b"cmd":
            assert content == b"STOP"
            self.exit()


class ProcessLoggingManager(QObject):
    """
    Receive and log logging messages from workers.

    An alternative might be using python logging's QueueListener, which
    like this code, runs on its own thread.
    """

    ready = pyqtSignal(int)

    @pyqtSlot()
    def startReceiver(self) -> None:
        context = zmq.Context.instance()
        self.receiver = context.socket(zmq.SUB)
        # Subscribe to all variates of logging messages
        self.receiver.setsockopt(zmq.SUBSCRIBE, b"")

        # Socket to receive subscription information, and the stop command
        info_socket = context.socket(zmq.PULL)
        self.info_port = info_socket.bind_to_random_port("tcp://*")

        poller = zmq.Poller()
        poller.register(self.receiver, zmq.POLLIN)
        poller.register(info_socket, zmq.POLLIN)

        self.ready.emit(self.info_port)

        while True:
            try:
                socks = dict(poller.poll())
            except KeyboardInterrupt:
                break

            if self.receiver in socks:
                message = self.receiver.recv()
                record = logging.makeLogRecord(pickle.loads(message))
                logger.handle(record)

            if info_socket in socks:
                directive, content = info_socket.recv_multipart()
                if directive == b"STOP":
                    break
                elif directive == b"CONNECT":
                    self.addSubscription(content)
                else:
                    assert directive == b"DISCONNECT"
                    self.removeSubscription(content)

    def addSubscription(self, port: bytes) -> None:
        try:
            port = int(port)
        except ValueError:
            logging.critical(
                "Incorrect port value in add logging subscription: %s", port
            )
        else:
            logging.debug("Subscribing to logging on port %s", port)
            self.receiver.connect(f"tcp://localhost:{port}")

    def removeSubscription(self, port: bytes):
        try:
            port = int(port)
        except ValueError:
            logging.critical(
                "Incorrect port value in remove logging subscription: %s", port
            )
        else:
            logging.debug("Unsubscribing to logging on port %s", port)
            self.receiver.disconnect(f"tcp://localhost:{port}")


def stop_process_logging_manager(info_port: int) -> None:
    """
    Stop ProcessLoggingManager thread

    :param info_port: the port number the manager uses
    """

    context = zmq.Context.instance()
    command = context.socket(zmq.PUSH)
    command.connect(f"tcp://localhost:{info_port}")
    command.send_multipart([b"STOP", b""])


class ScanArguments:
    """
    Pass arguments to the scan process
    """

    def __init__(
        self, device: Device, ignore_other_types: bool, log_gphoto2: bool
    ) -> None:
        """
        Pass arguments to the scan process

        :param device: the device to scan
        :param ignore_other_types: ignore file types like TIFF
        :param log_gphoto2: whether to generate detailed gphoto2 log
         messages
        """

        self.device = device
        self.ignore_other_types = ignore_other_types
        self.log_gphoto2 = log_gphoto2


class ScanResults:
    """
    Receive results from the scan process
    """

    def __init__(
        self,
        rpd_files: list[RPDFile] | None = None,
        file_type_counter: FileTypeCounter | None= None,
        file_size_sum: FileSizeSum | None= None,
        error_code: CameraErrorCode | None= None,
        error_message: str|None = None,
        scan_id: int|None = None,
        optimal_display_name: str|None = None,
        storage_space: list[StorageSpace] | None = None,
        storage_descriptions: list[str]|None = None,
        mount_point: str|None = None,
        is_apple_mobile: bool|None = False,
        sample_photo: Photo|None = None,
        sample_video: Video|None = None,
        problems: ScanProblems | None= None,
        fatal_error: bool|None = None,
        camera_removed: bool|None = None,
        entire_video_required: bool|None = None,
        entire_photo_required: bool|None = None,
    ) -> None:
        self.rpd_files = rpd_files
        self.file_type_counter = file_type_counter
        self.file_size_sum = file_size_sum
        self.error_code = error_code
        if error_code and error_message is None:
            self.error_message = ""
        else:
            self.error_message = error_message
        self.scan_id = scan_id
        self.optimal_display_name = optimal_display_name
        self.storage_space = storage_space
        self.storage_descriptions = storage_descriptions
        self.mount_point = mount_point
        self.is_apple_mobile = is_apple_mobile
        self.sample_photo = sample_photo
        self.sample_video = sample_video
        self.problems = problems
        self.fatal_error = fatal_error
        self.camera_removed = camera_removed
        self.entire_video_required = entire_video_required
        self.entire_photo_required = entire_photo_required


class CopyFilesArguments:
    """
    Pass arguments to the copyfiles process
    """

    def __init__(
        self,
        scan_id: int,
        device: Device,
        photo_download_folder: str,
        video_download_folder: str,
        files: list[RPDFile],
        verify_file: bool,
        generate_thumbnails: bool,
        log_gphoto2: bool,
    ) -> None:
        self.scan_id = scan_id
        self.device = device
        self.photo_download_folder = photo_download_folder
        self.video_download_folder = video_download_folder
        self.files = files
        self.generate_thumbnails = generate_thumbnails
        self.verify_file = verify_file
        self.log_gphoto2 = log_gphoto2


class CopyFilesResults:
    """
    Receive results from the copyfiles process
    """

    def __init__(
        self,
        scan_id: int|None = None,
        photo_temp_dir: str|None = None,
        video_temp_dir: str|None = None,
        total_downloaded: int|None = None,
        chunk_downloaded: int|None = None,
        copy_succeeded: bool|None = None,
        rpd_file: RPDFile | None = None,
        download_count: int|None = None,
        mdata_exceptions: tuple|None = None,
        problems: CopyingProblems | None = None,
        camera_removed: bool|None = None,
    ) -> None:
        """

        :param scan_id: scan id of the device the files are being
         downloaded from
        :param photo_temp_dir: temp directory path, used to copy
         photos into until they're renamed
        :param video_temp_dir: temp directory path, used to copy
         videos into until they're renamed
        :param total_downloaded: how many bytes in total have been
         downloaded
        :param chunk_downloaded: how many bytes were downloaded since
         the last message
        :param copy_succeeded: whether the copy was successful or not
        :param rpd_file: details of the file that was copied
        :param download_count: a running count of how many files
         have been copied. Used in download tracking.
        :param mdata_exceptions: details of errors setting file metadata
        :param problems: details of any problems encountered copying files,
         not including metedata write problems.
        """

        self.scan_id = scan_id

        self.photo_temp_dir = photo_temp_dir
        self.video_temp_dir = video_temp_dir

        self.total_downloaded = total_downloaded
        self.chunk_downloaded = chunk_downloaded

        self.copy_succeeded = copy_succeeded
        self.rpd_file = rpd_file
        self.download_count = download_count
        self.mdata_exceptions = mdata_exceptions
        self.problems = problems
        self.camera_removed = camera_removed


class ThumbnailDaemonData:
    """
    Pass arguments to the thumbnail daemon process.

    Occurs after a file is downloaded & renamed, and also
    after a file is backed up.
    """

    def __init__(
        self,
        frontend_port: int|None = None,
        rpd_file: RPDFile | None = None,
        write_fdo_thumbnail: bool|None = None,
        use_thumbnail_cache: bool|None = None,
        backup_full_file_names: list[str]|None = None,
        fdo_name: str|None = None,
        force_exiftool: bool|None = None,
    ) -> None:
        self.frontend_port = frontend_port
        self.rpd_file = rpd_file
        self.write_fdo_thumbnail = write_fdo_thumbnail
        self.use_thumbnail_cache = use_thumbnail_cache
        self.backup_full_file_names = backup_full_file_names
        self.fdo_name = fdo_name
        self.force_exiftool = force_exiftool


class RenameAndMoveFileData:
    """
    Pass arguments to the renameandmovefile process
    """

    def __init__(
        self,
        rpd_file: RPDFile = None,
        download_count: int = None,
        download_succeeded: bool = None,
        message: RenameAndMoveStatus = None,
    ) -> None:
        self.rpd_file = rpd_file
        self.download_count = download_count
        self.download_succeeded = download_succeeded
        self.message = message


class RenameAndMoveFileResults:
    def __init__(
        self,
        move_succeeded: bool = None,
        rpd_file: RPDFile = None,
        download_count: int = None,
        stored_sequence_no: int = None,
        downloads_today: list[str] = None,
        problems: RenamingProblems | None = None,
    ) -> None:
        self.move_succeeded = move_succeeded
        self.rpd_file = rpd_file
        self.download_count = download_count
        self.stored_sequence_no = stored_sequence_no
        self.downloads_today = downloads_today
        self.problems = problems


class OffloadData:
    def __init__(
        self,
        thumbnail_rows: Sequence[ThumbnailDataForProximity] | None = None,
        proximity_seconds: int = None,
        rpd_files: Sequence[RPDFile] | None = None,
        strip_characters: bool|None = None,
        folders_preview: FoldersPreview | None = None,
    ) -> None:
        self.thumbnail_rows = thumbnail_rows
        self.proximity_seconds = proximity_seconds
        self.rpd_files = rpd_files
        self.strip_characters = strip_characters
        self.folders_preview = folders_preview


class OffloadResults:
    def __init__(
        self,
        proximity_groups: TemporalProximityGroups | None = None,
        folders_preview: FoldersPreview | None = None,
    ) -> None:
        self.proximity_groups = proximity_groups
        self.folders_preview = folders_preview


class BackupArguments:
    """
    Pass start up data to the back up process
    """

    def __init__(self, path: str, device_name: str) -> None:
        self.path = path
        self.device_name = device_name


class BackupFileData:
    """
    Pass file data to the backup process
    """

    def __init__(
        self,
        rpd_file: RPDFile | None = None,
        move_succeeded: bool|None = None,
        do_backup: bool|None = None,
        path_suffix: str|None = None,
        backup_duplicate_overwrite: bool|None = None,
        verify_file: bool|None = None,
        download_count: int|None = None,
        save_fdo_thumbnail: int|None = None,
        message: BackupStatus | None = None,
    ) -> None:
        self.rpd_file = rpd_file
        self.move_succeeded = move_succeeded
        self.do_backup = do_backup
        self.path_suffix = path_suffix
        self.backup_duplicate_overwrite = backup_duplicate_overwrite
        self.verify_file = verify_file
        self.download_count = download_count
        self.save_fdo_thumbnail = save_fdo_thumbnail
        self.message = message


class BackupResults:
    def __init__(
        self,
        scan_id: int,
        device_id: int,
        total_downloaded: int|None = None,
        chunk_downloaded: int|None = None,
        backup_succeeded: bool|None = None,
        do_backup: bool|None = None,
        rpd_file: RPDFile | None = None,
        backup_full_file_name: str|None = None,
        mdata_exceptions: tuple|None = None,
        problems: BackingUpProblems | None = None,
    ) -> None:
        self.scan_id = scan_id
        self.device_id = device_id
        self.total_downloaded = total_downloaded
        self.chunk_downloaded = chunk_downloaded
        self.backup_succeeded = backup_succeeded
        self.do_backup = do_backup
        self.rpd_file = rpd_file
        self.backup_full_file_name = backup_full_file_name
        self.mdata_exceptions = mdata_exceptions
        self.problems = problems


class GenerateThumbnailsArguments:
    def __init__(
        self,
        scan_id: int,
        rpd_files: list[RPDFile],
        name: str,
        proximity_seconds: int,
        cache_dirs: CacheDirs,
        need_photo_cache_dir: bool,
        need_video_cache_dir: bool,
        frontend_port: int,
        log_gphoto2: bool,
        camera: str|None = None,
        port: str|None = None,
        is_mtp_device: bool|None = None,
        entire_video_required: bool|None = None,
        entire_photo_required: bool|None = None,
    ) -> None:
        """
        List of files for which thumbnails are to be generated.
        All files  are assumed to have the same scan id.
        :param scan_id: id of the scan
        :param rpd_files: files from which to extract thumbnails
        :param name: name of the device
        :param proximity_seconds: the time elapsed between consecutive
         shots that is used to prioritize the order of thumbnail
         generation
        :param cache_dirs: the location where the cache directories
         should be created
        :param need_photo_cache_dir: if True, must use cache dir
         to extract photo thumbnail
        :param need_video_cache_dir: if True, must use cache dir
         to extract video thumbnail
        :param frontend_port: port to use to send to load balancer's
         front end
        :param log_gphoto2: if True, log libgphoto2 logging messages
        :param camera: If the thumbnails are being downloaded from a
         camera, this is the name of the camera, else None
        :param port: If the thumbnails are being downloaded from a
         camera, this is the port of the camera, else None
        :param entire_video_required: if the entire video is required
         to extract the thumbnail
        :param entire_photo_required: if the entire photo is required
         to extract the thumbnail
        """

        self.rpd_files = rpd_files
        self.scan_id = scan_id
        self.name = name
        self.proximity_seconds = proximity_seconds
        self.cache_dirs = cache_dirs
        self.need_photo_cache_dir = need_photo_cache_dir
        self.need_video_cache_dir = need_video_cache_dir
        self.frontend_port = frontend_port
        if camera is not None:
            assert port is not None
            assert entire_video_required is not None
        self.camera = camera
        self.port = port
        self.is_mtp_device = is_mtp_device
        self.log_gphoto2 = log_gphoto2
        self.entire_video_required = entire_video_required
        self.entire_photo_required = entire_photo_required


class GenerateThumbnailsResults:
    def __init__(
        self,
        rpd_file: RPDFile | None = None,
        thumbnail_bytes: bytes|None = None,
        scan_id: int|None = None,
        cache_dirs: CacheDirs | None = None,
        camera_removed: bool|None = None,
    ) -> None:
        self.rpd_file = rpd_file
        # If thumbnail_bytes is None, there is no thumbnail
        self.thumbnail_bytes = thumbnail_bytes
        self.scan_id = scan_id
        self.cache_dirs = cache_dirs
        self.camera_removed = camera_removed


class ThumbnailExtractorArgument:
    def __init__(
        self,
        rpd_file: RPDFile,
        task: ExtractionTask,
        processing: set[ExtractionProcessing],
        full_file_name_to_work_on: str,
        secondary_full_file_name: str,
        exif_buffer: bytearray | None,
        thumbnail_bytes: bytes,
        use_thumbnail_cache: bool,
        file_to_work_on_is_temporary: bool,
        write_fdo_thumbnail: bool,
        send_thumb_to_main: bool,
        force_exiftool: bool,
    ) -> None:
        self.rpd_file = rpd_file
        self.task = task
        self.processing = processing
        self.full_file_name_to_work_on = full_file_name_to_work_on
        self.secondary_full_file_name = secondary_full_file_name
        self.file_to_work_on_is_temporary = file_to_work_on_is_temporary
        self.exif_buffer = exif_buffer
        self.thumbnail_bytes = thumbnail_bytes
        self.use_thumbnail_cache = use_thumbnail_cache
        self.write_fdo_thumbnail = write_fdo_thumbnail
        self.send_thumb_to_main = send_thumb_to_main
        self.force_exiftool = force_exiftool


class RenameMoveFileManager(PushPullDaemonManager):
    """
    Manages the single instance daemon process that renames and moves
    files that have just been downloaded
    """

    message = pyqtSignal(bool, RPDFile, int)
    sequencesUpdate = pyqtSignal(int, list)
    renameProblems = pyqtSignal("PyQt_PyObject")

    def __init__(self, logging_port: int) -> None:
        super().__init__(logging_port=logging_port, thread_name=ThreadNames.rename)
        self._process_name = "Rename and Move File Manager"
        self._process_to_run = "renameandmovefile.py"

    def process_sink_data(self):
        data = pickle.loads(self.content)  # type: RenameAndMoveFileResults
        if data.move_succeeded is not None:

            self.message.emit(data.move_succeeded, data.rpd_file, data.download_count)

        elif data.problems is not None:
            self.renameProblems.emit(data.problems)
        else:
            assert data.stored_sequence_no is not None
            assert data.downloads_today is not None
            assert isinstance(data.downloads_today, list)
            self.sequencesUpdate.emit(data.stored_sequence_no, data.downloads_today)


class ThumbnailDaemonManager(PushPullDaemonManager):
    """
    Manages the process that extracts thumbnails after the file
    has already been downloaded and that writes FreeDesktop.org
    thumbnails. Not to be confused with ThumbnailManagerPara, which
    manages thumbnailing using processes that run in parallel,
    one for each device.
    """

    message = pyqtSignal(RPDFile, QPixmap)

    def __init__(self, logging_port: int) -> None:
        super().__init__(
            logging_port=logging_port, thread_name=ThreadNames.thumbnail_daemon
        )
        self._process_name = "Thumbnail Daemon Manager"
        self._process_to_run = "thumbnaildaemon.py"

    def process_sink_data(self) -> None:
        data = pickle.loads(self.content)  # type: GenerateThumbnailsResults
        if data.thumbnail_bytes is None:
            thumbnail = QPixmap()
        else:
            thumbnail = QImage.fromData(data.thumbnail_bytes)
            if thumbnail.isNull():
                thumbnail = QPixmap()
            else:
                thumbnail = QPixmap.fromImage(thumbnail)
        self.message.emit(data.rpd_file, thumbnail)


class OffloadManager(PushPullDaemonManager):
    """
    Handles tasks best run in a separate process
    """

    message = pyqtSignal(TemporalProximityGroups)
    downloadFolders = pyqtSignal(FoldersPreview)

    def __init__(self, logging_port: int) -> None:
        super().__init__(logging_port=logging_port, thread_name=ThreadNames.offload)
        self._process_name = "Offload Manager"
        self._process_to_run = "offload.py"

    def process_sink_data(self) -> None:
        data = pickle.loads(self.content)  # type: OffloadResults
        if data.proximity_groups is not None:
            self.message.emit(data.proximity_groups)
        elif data.folders_preview is not None:
            self.downloadFolders.emit(data.folders_preview)


class ScanManager(PublishPullPipelineManager):
    """
    Handles the processes that scan devices (cameras, external devices,
    this computer path)
    """

    scannedFiles = pyqtSignal(
        "PyQt_PyObject", "PyQt_PyObject", FileTypeCounter, "PyQt_PyObject", bool, bool
    )
    deviceError = pyqtSignal(int, CameraErrorCode, str)
    deviceDetails = pyqtSignal(int, "PyQt_PyObject", "PyQt_PyObject", str, str, bool)
    scanProblems = pyqtSignal(int, "PyQt_PyObject")
    fatalError = pyqtSignal(int)
    cameraRemovedDuringScan = pyqtSignal(int)

    def __init__(self, logging_port: int) -> None:
        super().__init__(logging_port=logging_port, thread_name=ThreadNames.scan)
        self._process_name = "Scan Manager"
        self._process_to_run = "scan.py"

    def process_sink_data(self) -> None:
        data = pickle.loads(self.content)  # type: ScanResults
        if data.rpd_files is not None:
            assert data.file_type_counter
            assert data.file_size_sum
            assert data.entire_video_required is not None
            assert data.entire_photo_required is not None
            self.scannedFiles.emit(
                data.rpd_files,
                (data.sample_photo, data.sample_video),
                data.file_type_counter,
                data.file_size_sum,
                data.entire_video_required,
                data.entire_photo_required,
            )
        else:
            assert data.scan_id is not None
            if data.error_code is not None:
                assert data.error_message is not None
                self.deviceError.emit(data.scan_id, data.error_code, data.error_message)
            elif data.optimal_display_name is not None or data.mount_point is not None:
                # Some values might have value None, so give them default value in case
                self.deviceDetails.emit(
                    data.scan_id,
                    data.storage_space or [],
                    data.storage_descriptions or [],
                    data.optimal_display_name,
                    data.mount_point or "",
                    data.is_apple_mobile or False,
                )
            elif data.problems is not None:
                self.scanProblems.emit(data.scan_id, data.problems)
            elif data.camera_removed is not None:
                self.cameraRemovedDuringScan.emit(data.scan_id)
            else:
                assert data.fatal_error
                self.fatalError.emit(data.scan_id)


class BackupManager(PublishPullPipelineManager):
    """
    Each backup "device" (it could be an external drive, or a user-
    specified path on the local file system) has associated with it one
    worker process. For example if photos and videos are both being
    backed up to the same external hard drive, one worker process
    handles both the photos and the videos. However if photos are being
    backed up to one drive, and videos to another, there would be a
    worker process for each drive (2 in total).
    """

    message = pyqtSignal(int, bool, bool, RPDFile, str, "PyQt_PyObject")
    bytesBackedUp = pyqtSignal("PyQt_PyObject", "PyQt_PyObject")
    backupProblems = pyqtSignal(int, "PyQt_PyObject")

    def __init__(self, logging_port: int) -> None:
        super().__init__(logging_port=logging_port, thread_name=ThreadNames.backup)
        self._process_name = "Backup Manager"
        self._process_to_run = "backupfile.py"

    def process_sink_data(self) -> None:
        data = pickle.loads(self.content)  # type: BackupResults
        if data.total_downloaded is not None:
            assert data.scan_id is not None
            assert data.chunk_downloaded >= 0
            assert data.total_downloaded >= 0
            self.bytesBackedUp.emit(data.scan_id, data.chunk_downloaded)
        elif data.backup_succeeded is not None:
            assert data.do_backup is not None
            assert data.rpd_file is not None
            self.message.emit(
                data.device_id,
                data.backup_succeeded,
                data.do_backup,
                data.rpd_file,
                data.backup_full_file_name,
                data.mdata_exceptions,
            )
        else:
            assert data.problems is not None
            self.backupProblems.emit(data.device_id, data.problems)


class CopyFilesManager(PublishPullPipelineManager):
    """
    Manage the processes that copy files from devices to the computer
    during the download process
    """

    message = pyqtSignal(bool, RPDFile, int, "PyQt_PyObject")
    tempDirs = pyqtSignal(int, str, str)
    bytesDownloaded = pyqtSignal(int, "PyQt_PyObject", "PyQt_PyObject")
    copyProblems = pyqtSignal(int, "PyQt_PyObject")
    cameraRemoved = pyqtSignal(int)

    def __init__(self, logging_port: int) -> None:
        super().__init__(logging_port=logging_port, thread_name=ThreadNames.copy)
        self._process_name = "Copy Files Manager"
        self._process_to_run = "copyfiles.py"

    def process_sink_data(self) -> None:
        data = pickle.loads(self.content)  # type: CopyFilesResults
        if data.total_downloaded is not None:
            assert data.scan_id is not None
            if data.chunk_downloaded < 0:
                logging.critical(
                    "Chunk downloaded is less than zero: %s", data.chunk_downloaded
                )
            if data.total_downloaded < 0:
                logging.critical(
                    "Chunk downloaded is less than zero: %s", data.total_downloaded
                )

            self.bytesDownloaded.emit(
                data.scan_id, data.total_downloaded, data.chunk_downloaded
            )

        elif data.copy_succeeded is not None:
            assert data.rpd_file is not None
            assert data.download_count is not None
            self.message.emit(
                data.copy_succeeded,
                data.rpd_file,
                data.download_count,
                data.mdata_exceptions,
            )

        elif data.problems is not None:
            self.copyProblems.emit(data.scan_id, data.problems)
        elif data.camera_removed is not None:
            self.cameraRemoved.emit(data.scan_id)

        else:
            assert data.photo_temp_dir is not None and data.video_temp_dir is not None
            assert data.scan_id is not None
            self.tempDirs.emit(data.scan_id, data.photo_temp_dir, data.video_temp_dir)



# Copyright (C) 2015-2024 Damon Lynch <damonlynch@gmail.com>
# Copyright (C) 2008-2015 Canonical Ltd.
# Copyright (C) 2013 Bernard Baeyens

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
The primary task of this module is to handle addition and removal of
(1) cameras and (2) devices with file systems.

There are two scenarios:

1) User is running under a Gnome-like environment in which GVFS will
automatically mount cameras and devices. We can monitor mounts and
send a signal when something is mounted. The camera must be
unmounted before libgphoto2 can access it, so we must handle that too.

2) User is running under a non Gnome-like environment (e.g. KDE) in
which GVFS may or may not be running. However we can assume GVFS will
not automatically mount cameras and devices. In this case, using GIO
to monitor mounts is useless, as the mounts may not occur. So we must
monitor when cameras and other devices are added or removed ourselves.
To do this, use udev for cameras, and udisks2 for devices with file
systems. When a device with a file system is inserted, if it is not
already mounted, attempt to mount it.

The secondary task of this module is to provide miscellaneous services
regarding mount points and XDG related functionality.
"""

__author__ = "Damon Lynch"
__copyright__ = (
    "Copyright 2011-2024, Damon Lynch. Copyright 2008-2015 Canonical Ltd. Copyright"
    " 2013 Bernard Baeyens."
)

# ruff: noqa: E402


import functools
import logging
import os
import pwd
import re
import shutil
import sys
import time
from collections import namedtuple
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import NamedTuple
from urllib.parse import quote
from urllib.request import pathname2url

import gi
from PyQt5.QtCore import (
    QFileSystemWatcher,
    QObject,
    QStandardPaths,
    QStorageInfo,
    QTimer,
    pyqtSignal,
    pyqtSlot,
)
from showinfm import LinuxDesktop, linux_desktop, valid_file_manager
from tenacity import retry, stop_after_attempt, wait_fixed

from raphodo.wslutils import (
    wsl_conf_mnt_location,
    wsl_filter_directories,
    wsl_home,
    wsl_pictures_folder,
    wsl_videos_folder,
)

gi.require_version("GUdev", "1.0")
gi.require_version("UDisks", "2.0")
gi.require_version("GExiv2", "0.10")
gi.require_version("GLib", "2.0")
from gi.repository import GLib, GUdev, UDisks

from raphodo.constants import Distro, PostCameraUnmountAction
from raphodo.utilities import (
    log_os_release,
    remove_topmost_directory_from_path,
)

logging_level = logging.DEBUG

try:
    from gi.repository import Gio

    have_gio = True
except ImportError:
    have_gio = False

StorageSpace = namedtuple("StorageSpace", "bytes_free, bytes_total, path")
CameraDetails = namedtuple(
    "CameraDetails", "model, port, display_name, is_mtp, storage_desc"
)
UdevAttr = namedtuple(
    "UdevAttr", "is_mtp_device, vendor, model, is_apple_mobile, serial"
)

PROGRAM_DIRECTORY = "rapid-photo-downloader"

try:
    _linux_desktop = linux_desktop()
except Exception:
    _linux_desktop = LinuxDesktop.unknown


def guess_distro() -> Distro:
    """
    Guess distro support by checking package manager support
    :return:
    """

    if shutil.which("apt") or shutil.which("apt-get"):
        return Distro.debian_derivative
    if shutil.which("dnf"):
        return Distro.fedora_derivative
    return Distro.unknown


def parse_os_release() -> dict[str, str]:
    d = {}
    if os.path.isfile("/etc/os-release"):
        with open("/etc/os-release") as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith("#"):
                    kv = line.split("=", maxsplit=1)
                    if len(kv) == 2:
                        k, v = kv
                        v = v.strip("'\"")
                        d[k] = v
    return d


def get_distro() -> Distro:
    """
    Determine the Linux distribution using /etc/os-release
    :param os_release: parsed /etc/os-release file
    """

    os_release = parse_os_release()
    name = os_release.get("NAME")

    distro = None

    if name:
        if "Ubuntu" in name:
            distro = Distro.ubuntu
        if "Fedora" in name:
            distro = Distro.fedora
        if "CentOS Linux" in name:
            version_id = os_release.get("VERSION_ID")
            distro = Distro.centos7 if version_id == "7" else Distro.centos8
        if "CentOS Stream" in name:
            version_id = os_release.get("VERSION_ID")
            if version_id == "8":
                distro = Distro.centos_stream8
            else:
                distro = Distro.centos_stream9
        if "Linux Mint" in name:
            distro = Distro.linuxmint
        if "elementary" in name:
            distro = Distro.elementary
        if "openSUSE" in name:
            distro = Distro.opensuse
        if "Deepin" in name:
            distro = Distro.deepin
        if "KDE neon" in name:
            distro = Distro.neon
        if "Zorin" in name:
            distro = Distro.zorin
        if "Kylin" in name:
            distro = Distro.kylin
        if "Pop!_OS" in name:
            distro = Distro.popos
        if "Raspbian" in name:
            distro = Distro.raspbian
        if "Debian" in name:
            distro = Distro.debian
        if "Manjaro" in name:
            distro = Distro.manjaro
        if "Gentoo" in name:
            distro = Distro.gentoo

    if distro is None:
        idlike = os_release.get("ID_LIKE")
        if idlike:
            if "arch" in idlike:
                distro = Distro.arch
            if "ubuntu" in idlike:
                distro = Distro.ubuntu_derivative
            if "debian" in idlike:
                distro = Distro.debian_derivative
            if "fedora" in idlike:
                distro = Distro.fedora_derivative

    if distro is None:
        distro = guess_distro()

    return distro


def get_user_name() -> str:
    """
    Gets the user name of the process owner, with no exception checking
    :return: user name of the process owner
    """

    return pwd.getpwuid(os.getuid())[0]


def get_path_display_name(path: str) -> tuple[str, str]:
    """
    Return a name for the path (path basename),
    removing a final '/' when it's not the root of the
    file system.

    :param path: path to generate the display name for
    :return: display name and sanitized path
    """
    if path.endswith(os.sep) and path != os.sep:
        path = path[:-1]

    display_name = _("File system root") if path == os.sep else os.path.basename(path)
    return display_name, path


@functools.cache
def get_media_dir() -> str:
    """
    Returns the media directory, i.e. where external mounts are mounted.

    Assumes mount point of /media/<USER>.

    """

    if sys.platform.startswith("linux"):
        if _linux_desktop == LinuxDesktop.wsl2:
            return wsl_conf_mnt_location()

        media_dir = f"/media/{get_user_name()}"
        run_media_dir = "/run/media"
        distro = get_distro()
        if os.path.isdir(run_media_dir) and distro not in (
            Distro.ubuntu,
            Distro.debian,
            Distro.neon,
            Distro.galliumos,
            Distro.peppermint,
            Distro.elementary,
            Distro.zorin,
            Distro.popos,
        ):
            if distro not in (
                Distro.fedora,
                Distro.manjaro,
                Distro.arch,
                Distro.opensuse,
                Distro.gentoo,
                Distro.centos8,
                Distro.centos_stream8,
                Distro.centos_stream9,
                Distro.centos7,
            ):
                logging.debug(
                    "Detected /run/media directory, but distro does not appear "
                    "to be CentOS, Fedora, Arch, openSUSE, Gentoo, or Manjaro"
                )
                log_os_release()
            return run_media_dir
        return media_dir
    else:
        raise ("Mounts.setValidMountPoints() not implemented on %s", sys.platform)


_gvfs_gphoto2 = re.compile("gvfs.*gphoto2.*host")


def gvfs_gphoto2_path(path: str) -> bool:
    """
    :return: True if the path appears to be a GVFS gphoto2 path

    >>> p = "/run/user/1000/gvfs/gphoto2:host=%5Busb%3A002%2C013%5D"
    >>> gvfs_gphoto2_path(p)
    True
    >>> p = '/home/damon'
    >>> gvfs_gphoto2_path(p)
    False
    """

    return _gvfs_gphoto2.search(path) is not None


class ValidMounts:
    r"""
    Operations to find 'valid' mount points, i.e. the places in which
    it's sensible for a user to mount a partition. Valid mount points:
    include /home/<USER> , /media/<USER>, and /run/media/<USER>
    include directories in /etc/fstab, except /, /home, and swap
    However if only considering external mounts, the the mount must be
    under /media/<USER> or /run/media/<user>
    """

    def __init__(self, only_external_mounts: bool):
        """
        :param only_external_mounts: if True, valid mounts must be under
        /media/<USER>, /run/media/<user>, or if WSL2 /mnt/
        """
        self.validMountFolders = None  # type: tuple[str, ...] | None
        self.only_external_mounts = only_external_mounts
        self.is_wsl2 = _linux_desktop == LinuxDesktop.wsl2
        self._setValidMountFolders()
        assert "/" not in self.validMountFolders
        if logging_level == logging.DEBUG:
            self.logValidMountFolders()

    def isValidMountPoint(self, mount: QStorageInfo) -> bool:
        """
        Determine if the path of the mount point starts with a valid
        path
        :param mount: QStorageInfo to be tested
        :return:True if mount is a mount under a valid mount, else False
        """
        root_path = mount.rootPath()
        if self.is_wsl2:
            for path in wsl_filter_directories():
                if root_path.startswith(path):
                    return False
        return any(root_path.startswith(m) for m in self.validMountFolders)

    def pathIsValidMountPoint(self, path: str) -> bool:
        """
        Determine if path indicates a mount point under a valid mount
        point
        :param path: path to be tested
        :return:True if path is a mount under a valid mount, else False
        """
        return any(path.startswith(m) for m in self.validMountFolders)

    def mountedValidMountPointPaths(self) -> tuple[str]:
        """
        Return paths of all the currently mounted partitions that are
        valid
        :return: tuple of currently mounted valid partition paths
        """

        return tuple(filter(self.pathIsValidMountPoint, mountPaths()))

    def mountedValidMountPoints(self) -> tuple[QStorageInfo]:
        """
        Return mount points of all the currently mounted partitions
        that are valid
        :return: tuple of currently mounted valid partition
        """

        return tuple(filter(self.isValidMountPoint, QStorageInfo.mountedVolumes()))

    def _setValidMountFolders(self) -> None:
        """
        Determine the valid mount point folders and set them in
        self.validMountFolders, e.g. /media/<USER>, etc.
        """

        if not sys.platform.startswith("linux"):
            raise ("Mounts.setValidMountPoints() not implemented on %s", sys.platform)
        else:
            try:
                media_dir = get_media_dir()
            except Exception:
                logging.critical("Unable to determine username of this process")
                media_dir = ""
            logging.debug("Media dir is %s", media_dir)
            if self.only_external_mounts:
                self.validMountFolders = (media_dir,)
            else:
                home_dir = os.path.expanduser("~")
                validPoints = [home_dir, media_dir]
                for point in self.mountPointInFstab():
                    validPoints.append(point)
                self.validMountFolders = tuple(validPoints)

    def mountPointInFstab(self):
        """
        Yields a list of mount points in /etc/fstab
        The mount points will exclude /, /home, and swap
        """

        with open("/etc/fstab") as f:
            for line in f:
                # As per fstab specs: white space is either Tab or space
                # Ignore comments, blank lines
                # Also ignore swap file (mount point none), root, and /home
                m = re.match(
                    r"^(?![\t ]*#)\S+\s+(?!(none|/[\t ]|/home))(" r"?P<point>\S+)", line
                )
                if m is not None:
                    yield (m.group("point"))

    def logValidMountFolders(self):
        """
        Output nicely formatted debug logging message
        """

        assert len(self.validMountFolders) > 0
        if logging_level == logging.DEBUG:
            msg = "To be recognized, partitions must be mounted under "
            if len(self.validMountFolders) > 2:
                msg += "one of "
                for p in self.validMountFolders[:-2]:
                    msg += f"{p}, "
                msg += f"{self.validMountFolders[-2]} or {self.validMountFolders[-1]}"
            elif len(self.validMountFolders) == 2:
                msg += f"{self.validMountFolders[0]} or {self.validMountFolders[1]}"
            else:
                msg += self.validMountFolders[0]
            logging.debug(msg)


def mountPaths():
    """
    Yield all the mount paths returned by QStorageInfo
    """

    for m in QStorageInfo.mountedVolumes():
        yield m.rootPath()


def has_one_or_more_folders(path: str, folders: list[str]) -> bool:
    """
    Checks to see if directly below the path there is a folder
    from the list of specified folders, and if the folder is readable.
    :param path: path to check
    :return: True if has one or more valid folders, False otherwise
    """

    try:
        contents = os.listdir(path)
        for folder in folders:
            if folder in contents:
                full_path = os.path.join(path, folder)
                if os.path.isdir(full_path) and os.access(full_path, os.R_OK):
                    return True
    except (PermissionError, FileNotFoundError, OSError):
        return False
    except Exception:
        logging.error(
            "Unknown error occurred while probing potential source folder %s", path
        )
        return False
    return False


def get_desktop_environment() -> str | None:
    """
    Determine desktop environment using environment variable XDG_CURRENT_DESKTOP

    :return: str with XDG_CURRENT_DESKTOP value
    """

    return os.getenv("XDG_CURRENT_DESKTOP")


def _platform_special_dir(
    dir_type: QStandardPaths, home_on_failure: bool = True
) -> str | None:
    """
    Use Qt to query the platforms standard paths

    :param dir_type: one of Qt's standard paths
    :param home_on_failure: return the home directory if the special path cannot
     be located
    :return: the directory, or None if it cannot be determined
    """

    path = QStandardPaths.writableLocation(dir_type)
    if path:
        return path
    elif home_on_failure:
        try:
            return str(Path.home())
        except RuntimeError:
            logging.error("Unable to determine home directory")
    return None


def platform_photos_directory(home_on_failure: bool = True) -> str | None:
    """
    Get localized version of /home/<USER>/Pictures

    :param home_on_failure: if the directory does not exist, return
     the home directory instead
    :return: the directory if it is specified, else the user's
    home directory or None
    """

    if _linux_desktop == LinuxDesktop.wsl2:
        try:
            return wsl_pictures_folder()
        except Exception as e:
            logging.error("Error querying Windows registry: %s", str(e))
        path = wsl_home()
        if path.is_dir():
            return str(path / "Pictures")
    return _platform_special_dir(QStandardPaths.PicturesLocation, home_on_failure)


def platform_videos_directory(home_on_failure: bool = True) -> str:
    """
    Get localized version of /home/<USER>/Videos

    :param home_on_failure: if the directory does not exist, return
     the home directory instead
    :return: the directory if it is specified, else the user's
    home directory or None
    """

    if _linux_desktop == LinuxDesktop.wsl2:
        try:
            return wsl_videos_folder()
        except Exception as e:
            logging.error("Error querying Windows registry: %s", str(e))
        path = wsl_home()
        if path.is_dir():
            return str(path / "Videos")
    return _platform_special_dir(QStandardPaths.MoviesLocation, home_on_failure)


def platform_desktop_directory(home_on_failure: bool = True) -> str:
    """
    Get localized version of /home/<USER>/Desktop

    :param home_on_failure: if the directory does not exist, return
     the home directory instead
    :return: the directory if it is specified, else the user's
    home directory or None
    """
    return _platform_special_dir(QStandardPaths.DesktopLocation, home_on_failure)


def platform_photos_identifier() -> str:
    """
    Get special subfoler indicated by the localized version of /home/<USER>/Pictures
    :return: the subfolder name if it is specified, else the localized version of
    'Pictures'
    """

    path = _platform_special_dir(QStandardPaths.PicturesLocation, home_on_failure=False)
    if path is None:
        # translators: the name of the Pictures folder
        return _("Pictures")
    return os.path.basename(path)


def platform_videos_identifier() -> str:
    """
    Get special subfoler indicated by the localized version of /home/<USER>/Pictures
    :return: the subfolder name if it is specified, else the localized version of
    'Pictures'
    """

    path = _platform_special_dir(QStandardPaths.MoviesLocation, home_on_failure=False)
    if path is None:
        # translators: the name of the Videos folder
        return _("Videos")
    return os.path.basename(path)


def make_program_directory(path: str) -> str:
    """
    Creates a subfolder used by Rapid Photo Downloader.

    Does not catch errors.

    :param path: location where the subfolder should be
    :return: the full path of the new directory
    """

    program_dir = os.path.join(path, "rapid-photo-downloader")
    if not os.path.exists(program_dir):
        os.mkdir(program_dir)
    elif not os.path.isdir(program_dir):
        os.remove(program_dir)
        os.mkdir(program_dir)
    return program_dir


def get_program_cache_directory(create_if_not_exist: bool = False) -> str | None:
    """
    Get Rapid Photo Downloader cache directory.

    :param create_if_not_exist: creates directory if it does not exist.
    :return: the full path of the cache directory, or None on error
    """

    # Must use GenericCacheLocation, never CacheLocation
    cache_directory = _platform_special_dir(
        QStandardPaths.GenericCacheLocation, home_on_failure=False
    )
    if cache_directory is None:
        logging.error("The platform's cache directory could not be determined")
        return None
    try:
        if not create_if_not_exist:
            return str(Path(cache_directory) / PROGRAM_DIRECTORY)
        else:
            return make_program_directory(cache_directory)
    except OSError:
        logging.error("An error occurred while creating the cache directory")
        return None


def get_program_logging_directory(create_if_not_exist: bool = False) -> str | None:
    """
    Get directory in which to store program log files.

    Log files are kept in the cache directory.

    :param create_if_not_exist: create the directory if it does not exist
    :return: the full path of the logging directory, or None on error
    """

    cache_directory = get_program_cache_directory(
        create_if_not_exist=create_if_not_exist
    )
    if cache_directory is None:
        logging.error("Unable to create logging directory")
        return None
    log_dir = os.path.join(cache_directory, "log")
    if os.path.isdir(log_dir):
        return log_dir
    if create_if_not_exist:
        try:
            if os.path.isfile(log_dir):
                os.remove(log_dir)
            os.mkdir(log_dir, 0o700)
            return log_dir
        except OSError:
            logging.error("An error occurred while creating the log directory")
    return None


def get_program_data_directory(create_if_not_exist=False) -> str | None:
    """
    Get Rapid Photo Downloader data directory, which is assumed to be
    under $XDG_DATA_HOME or if that doesn't exist,  ~/.local/share
    :param create_if_not_exist: creates directory if it does not exist.
    :return: the full path of the data directory, or None on error
    """

    data_directory = _platform_special_dir(
        QStandardPaths.GenericDataLocation, home_on_failure=False
    )
    if data_directory is None:
        logging.error("The program's data directory could not be determined")
        return None
    if not create_if_not_exist:
        return str(Path(data_directory) / PROGRAM_DIRECTORY)
    else:
        return make_program_directory(data_directory)


def get_fdo_cache_thumb_base_directory() -> str:
    """
    Get the Freedesktop.org thumbnail directory location
    :return: location
    """

    # LXDE is a special case: handle it
    if _linux_desktop == LinuxDesktop.lxde:
        return str(Path.home() / ".thumbnails")

    cache = _platform_special_dir(
        QStandardPaths.GenericCacheLocation, home_on_failure=False
    )
    try:
        return str(Path(cache) / "thumbnails")
    except TypeError:
        logging.error("Could not determine freedesktop.org thumbnail cache location")
        raise "Could not determine freedesktop.org thumbnail cache location"


# Module level variables important for determining among other things the generation of
# URIs
_quoted_comma = quote(",")
_valid_file_manager_probed = False
_valid_file_manager = None  # type: str|None

gvfs_file_managers = (
    "nautilus",
    "caja",
    "thunar",
    "nemo",
    "pcmanfm",
    "peony",
    "pcmanfm-qt",
    "dde-file-manager",
    "io.elementary.files",
)

kframework_file_managers = ("dolphin", "index", "krusader")


def get_uri(
    full_file_name: str | None = None,
    path: str | None = None,
    camera_details: CameraDetails | None = None,
) -> str:
    """
    Generate and return the URI for the file, which varies depending on
    which device the file is located

    :param full_file_name: full filename and path
    :param path: straight path when not passing a full_file_name
    :param camera_details: see named tuple CameraDetails for parameters
    :param desktop_environment: if True, will to generate a URI accepted
     by Gnome, KDE and other desktops, which means adjusting the URI if it appears to
     be an MTP mount. Includes the port too, for cameras. Takes into account
     file manager characteristics.
    :return: the URI
    """

    global _valid_file_manager
    global _valid_file_manager_probed
    if not _valid_file_manager_probed:
        _valid_file_manager = valid_file_manager()
        _valid_file_manager_probed = True

    if camera_details is None:
        prefix = "file://"
    else:
        prefix = ""
        # Attempt to generate a URI accepted by desktop environments
        if camera_details.is_mtp:
            if full_file_name:
                full_file_name = remove_topmost_directory_from_path(full_file_name)
            elif path:
                path = remove_topmost_directory_from_path(path)

            if _valid_file_manager in gvfs_file_managers:
                prefix = "mtp://" + pathname2url(
                    f"[{camera_details.port}]/{camera_details.storage_desc}"
                )
            elif _valid_file_manager in kframework_file_managers:
                prefix = "mtp:/" + pathname2url(
                    f"{camera_details.display_name}/{camera_details.storage_desc}"
                )
            else:
                logging.error(
                    "Don't know how to generate MTP prefix for %s", _valid_file_manager
                )
        else:
            if _valid_file_manager in kframework_file_managers:
                prefix = (
                    "camera:/"
                    f"{pathname2url(camera_details.display_name.replace('-', ' '))}"
                    f"@{camera_details.port}"
                )
            else:
                prefix = "gphoto2://" + pathname2url(f"[{camera_details.port}]")

        if _valid_file_manager == "pcmanfm-qt":
            # pcmanfm-qt does not like the quoted form of the comma
            prefix = prefix.replace(_quoted_comma, ",")
            if full_file_name:
                # pcmanfm-qt does not like the filename as part of the path
                full_file_name = os.path.dirname(full_file_name)

    if full_file_name or path:
        uri = f"{prefix}{pathname2url(full_file_name or path)}"
    else:
        uri = prefix
    return uri


class ValidatedFolder(NamedTuple):
    valid: bool
    absolute_path: str


def validate_download_folder(
    path: str | None, write_on_waccesss_failure: bool = False
) -> ValidatedFolder:
    r"""
    Check if folder exists and is writeable.

    Accepts None as a folder, which will always be invalid.

    :param path: path to analyze
    :param write_on_waccesss_failure: if os.access reports path is not writable, test
     nonetheless to see if it's writable by writing and deleting a test file
    :return: Tuple indicating validity and path made absolute

    >>> validate_download_folder('/some/bogus/and/ridiculous/path')
    ValidatedFolder(valid=False, absolute_path='/some/bogus/and/ridiculous/path')
    >>> validate_download_folder(None)
    ValidatedFolder(valid=False, absolute_path='')
    >>> validate_download_folder('')
    ValidatedFolder(valid=False, absolute_path='')
    """

    if not path:
        return ValidatedFolder(valid=False, absolute_path="")
    absolute_path = os.path.abspath(path)
    valid = os.path.isdir(path) and os.access(path, os.W_OK)
    if not valid and write_on_waccesss_failure and os.path.isdir(path):
        try:
            with NamedTemporaryFile(dir=path):
                # the path is in fact writeable -- can happen with NFS
                valid = True
        except Exception:
            logging.warning(
                "While validating download / backup folder, failed to write a "
                "temporary file to %s",
                path,
            )

    return ValidatedFolder(valid=valid, absolute_path=absolute_path)


def validate_source_folder(path: str | None) -> ValidatedFolder:
    r"""
    Check if folder exists and is readable.

    Accepts None as a folder, which will always be invalid.

    :param path: path to analyze
    :return: Tuple indicating validity and path made absolute

    >>> validate_source_folder('/some/bogus/and/ridiculous/path')
    ValidatedFolder(valid=False, absolute_path='/some/bogus/and/ridiculous/path')
    >>> validate_source_folder(None)
    ValidatedFolder(valid=False, absolute_path='')
    >>> validate_source_folder('')
    ValidatedFolder(valid=False, absolute_path='')
    """

    if not path:
        return ValidatedFolder(valid=False, absolute_path="")
    absolute_path = os.path.abspath(path)
    valid = os.path.isdir(path) and os.access(path, os.R_OK)
    return ValidatedFolder(valid=valid, absolute_path=absolute_path)


def udev_attributes(devname: str) -> UdevAttr | None:
    """
    Query udev to see if a device is an MTP device.

    :param devname: udev DEVNAME e.g. '/dev/bus/usb/001/003'
    :return True if udev property ID_MTP_DEVICE == '1', else False
    """

    client = GUdev.Client(subsystems=["usb", "block"])
    enumerator = GUdev.Enumerator.new(client)
    enumerator.add_match_property("DEVNAME", devname)
    for device in enumerator.execute():
        model = device.get_property("ID_MODEL")  # type: str
        if model is not None:
            is_mtp = (
                device.get_property("ID_MTP_DEVICE") == "1"
                or device.get_property("ID_MEDIA_PLAYER") == "1"
            )
            vendor = device.get_property("ID_VENDOR")  # type: str
            model = model.replace("_", " ").strip()
            vendor = vendor.replace("_", " ").strip()

            is_apple_mobile = False
            if device.has_sysfs_attr("configuration"):
                config = device.get_sysfs_attr("configuration")
                if config is not None:
                    is_apple_mobile = config.lower().find("apple mobile") >= 0

            if not is_apple_mobile and vendor.lower().find("apple") >= 0:
                logging.warning(
                    "Setting Apple device detected to True even though Apple Mobile "
                    "UDEV configuration not set because vendor is %s",
                    vendor,
                )
                is_apple_mobile = True

            if device.has_sysfs_attr("serial"):
                serial = device.get_sysfs_attr("serial")
                logging.debug("Device serial: %s", serial)
            else:
                serial = None

            if is_apple_mobile:
                if serial:
                    logging.debug(
                        "Detected using udev Apple Mobile device at %s with serial %s",
                        devname,
                        serial,
                    )
                else:
                    logging.warning(
                        "Detected using udev Apple Mobile device at %s but could not "
                        "determine serial number",
                        devname,
                    )

            return UdevAttr(is_mtp, vendor, model, is_apple_mobile, serial)
    return None


def udev_is_camera(devname: str) -> bool:
    """
    Query udev to see if device is a gphoto2 device (a camera or phone)
    :param devname: udev DEVNAME e.g. '/dev/bus/usb/001/003'
    :return: True if so, else False
    """

    client = GUdev.Client(subsystems=["usb", "block"])
    enumerator = GUdev.Enumerator.new(client)
    enumerator.add_match_property("DEVNAME", devname)
    for device in enumerator.execute():
        if device.get_property("ID_GPHOTO2") == "1":
            return True
    return False


def fs_device_details(path: str) -> tuple[str, str, str, str]:
    """
    :return: device (volume) name, uri, root path and filesystem type
     of the mount the path is on
    """
    qsInfo = QStorageInfo(path)
    name = qsInfo.displayName()
    root_path = qsInfo.rootPath()
    uri = f"file://{pathname2url(root_path)}"
    fstype = qsInfo.fileSystemType()
    if isinstance(fstype, bytes):
        fstype = fstype.decode()
    return name, uri, root_path, fstype


class WatchDownloadDirs(QFileSystemWatcher):
    """
    Create a file system watch to monitor if there are changes to the
    download directories.

    Monitors the parent directory because we need to monitor it to detect if the
    download directory has been removed.
    """

    def updateWatchPathsFromPrefs(self, prefs) -> None:
        """
        Update the watched directories using values from the program preferences
        :param prefs: program preferences
        :type prefs: raphodo.preferences.Preferences
        """

        logging.debug("Updating watched paths")

        paths = (
            os.path.dirname(path)
            for path in (prefs.photo_download_folder, prefs.video_download_folder)
        )
        watch = {path for path in paths if path}

        existing_watches = set(self.directories())

        if watch == existing_watches:
            return

        new = watch - existing_watches
        if new:
            new = list(new)
            logging.debug("Adding to watched paths: %s", ", ".join(new))
            failures = self.addPaths(new)
            if failures:
                logging.debug("Failed to add watched paths: %s", failures)

        old = existing_watches - watch
        if old:
            old = list(old)
            logging.debug("Removing from watched paths: %s", ", ".join(old))
            failures = self.removePaths(old)
            if failures:
                logging.debug("Failed to remove watched paths: %s", failures)

    def closeWatch(self) -> None:
        """
        End all watches.
        """
        dirs = self.directories()
        if dirs:
            self.removePaths(dirs)


class CameraHotplug(QObject):
    cameraAdded = pyqtSignal()
    cameraRemoved = pyqtSignal()

    def __init__(self):
        super().__init__()
        self.cameras = {}

    @pyqtSlot()
    def startMonitor(self):
        self.client = GUdev.Client(subsystems=["usb", "block"])
        self.client.connect("uevent", self.ueventCallback)
        logging.debug("... camera hotplug monitor started")
        self.enumerateCameras()
        if self.cameras:
            logging.info(
                "Camera Hotplug found %d camera(s): %s",
                len(self.cameras),
                ", ".join((model for port, model in self.cameras.items())),
            )
            for port, model in self.cameras.items():
                logging.debug("%s is at %s", model, port)

    def enumerateCameras(self):
        """
        Query udev to get the list of cameras store their path and
        model in our internal dict, which is useful when responding to
        camera removal.
        """
        enumerator = GUdev.Enumerator.new(self.client)
        enumerator.add_match_property("ID_GPHOTO2", "1")
        for device in enumerator.execute():
            model = device.get_property("ID_MODEL")
            if model is not None:
                path = device.get_sysfs_path()
                self.cameras[path] = model

    def ueventCallback(
        self, client: GUdev.Client, action: str, device: GUdev.Device
    ) -> None:
        # for key in device.get_property_keys():
        #     print(key, device.get_property(key))
        if device.get_property("ID_GPHOTO2") == "1":
            self.camera(action, device)

    def camera(self, action: str, device: GUdev.Device) -> None:
        # For some reason, the add and remove camera event is triggered twice.
        # The second time the device information is a variation on information
        # from the first time.
        path = device.get_sysfs_path()
        parent_device = device.get_parent()
        parent_path = parent_device.get_sysfs_path()
        logging.debug(
            "Device change: %s. Path: %s Parent path: %s", action, path, parent_path
        )
        if device.has_property("ID_VENDOR_FROM_DATABASE"):
            vendor = device.get_property("ID_VENDOR_FROM_DATABASE")
            logging.debug("Device vendor: %s", vendor)
        else:
            vendor = ""

        # 'bind' vs 'add' action: see https://lwn.net/Articles/837033/

        if action == "bind":
            if parent_path not in self.cameras:
                model = ""

                if device.has_property("ID_MODEL"):
                    model = device.get_property("ID_MODEL")
                    model = model.replace("_", " ")
                    camera_path = path
                else:
                    camera_path = parent_path

                name = model or vendor or "unknown camera"

                logging.info("Hotplug: new camera: %s", name)
                self.cameras[camera_path] = name
                self.cameraAdded.emit()
            else:
                logging.debug(
                    "Hotplug: already know about %s", self.cameras[parent_path]
                )

        elif action == "remove":
            emit_remove = False
            name = ""

            # A path might look like:
            # /sys/devices/pci0000:00/0000:00:1c.6/0000:0e:00.0/usb3/3-2/3-2:1.0
            # When what we want is:
            # /sys/devices/pci0000:00/0000:00:1c.6/0000:0e:00.0/usb3/3-2
            # This unchanged path used to work, so test both the unchanged and modified
            # path
            # Note enumerateCameras() above finds only the path as in the 2nd type,
            # without the 3-2:1.0
            split_path = os.path.split(path)[0]

            for p in (path, split_path):
                if p in self.cameras:
                    name = self.cameras[p]
                    logging.debug("Hotplug: removing '%s' on basis of path %s", name, p)
                    del self.cameras[p]
                    emit_remove = True
                    break

            if emit_remove:
                logging.info("Hotplug: '%s' has been removed", name)
                self.cameraRemoved.emit()
            else:
                logging.debug(
                    "Not responding to device removal: '%s'",
                    vendor or device.get_sysfs_path(),
                )


class UDisks2Monitor(QObject):
    # Most of this class is Copyright 2008-2015 Canonical

    partitionMounted = pyqtSignal(str, "PyQt_PyObject", bool)
    partitionUnmounted = pyqtSignal(str)

    loop_prefix = "/org/freedesktop/UDisks2/block_devices/loop"
    not_interesting = (
        "/org/freedesktop/UDisks2/block_devices/dm_",
        "/org/freedesktop/UDisks2/block_devices/ram",
        "/org/freedesktop/UDisks2/block_devices/zram",
    )

    def __init__(self, validMounts: ValidMounts, prefs) -> None:
        super().__init__()
        self.prefs = prefs
        self.validMounts = validMounts

    @pyqtSlot()
    def startMonitor(self) -> None:
        self.udisks = UDisks.Client.new_sync(None)
        self.manager = self.udisks.get_object_manager()
        self.manager.connect(
            "object-added", lambda man, obj: self._udisks_obj_added(obj)
        )
        self.manager.connect(
            "object-removed", lambda man, obj: self._device_removed(obj)
        )

        # Track the paths of the mount points, which is useful when unmounting
        # objects.
        self.known_mounts = {}  # type: dict[str, str]
        for obj in self.manager.get_objects():
            path = obj.get_object_path()
            fs = obj.get_filesystem()
            if fs:
                mount_points = fs.get_cached_property(
                    "MountPoints"
                ).get_bytestring_array()
                if mount_points:
                    self.known_mounts[path] = mount_points[0]
        logging.debug("... UDisks2 monitor started")

    def _udisks_obj_added(self, obj) -> None:
        path = obj.get_object_path()
        for boring in self.not_interesting:
            if path.startswith(boring):
                return
        block = obj.get_block()
        if not block:
            return

        drive = self._get_drive(block)

        part = obj.get_partition()
        is_system = block.get_cached_property("HintSystem").get_boolean()
        is_loop = (
            path.startswith(self.loop_prefix)
            and not block.get_cached_property("ReadOnly").get_boolean()
        )
        if (not is_system or is_loop) and part:
            self._udisks_partition_added(obj, block, drive, path)

    def _get_drive(self, block) -> UDisks.Drive | None:
        drive_name = block.get_cached_property("Drive").get_string()
        if drive_name != "/":
            return self.udisks.get_object(drive_name).get_drive()
        else:
            return None

    def _udisks_partition_added(self, obj, block, drive, path) -> None:
        logging.debug("UDisks: partition added: %s" % path)
        fstype = block.get_cached_property("IdType").get_string()
        logging.debug("Udisks: id-type: %s" % fstype)

        fs = obj.get_filesystem()

        if fs:
            icon_names = self.get_icon_names(obj)

            ejectable = drive.get_property("ejectable") if drive is not None else False
            if not self.prefs.auto_mount:
                logging.debug(
                    "Not mounting device because auto mount preference is off: %s",
                    path,
                )
                return

            mount_points = fs.get_cached_property("MountPoints").get_bytestring_array()
            if len(mount_points) == 0:
                try:
                    logging.debug("UDisks: attempting to mount %s", path)
                    mount_point = self.retry_mount(fs, fstype)
                    if not mount_point:
                        raise Exception
                    else:
                        logging.debug("UDisks: successfully mounted at %s", mount_point)
                except Exception:
                    logging.error("UDisks: could not mount the device: %s", path)
                    return
            else:
                mount_point = mount_points[0]
                logging.debug("UDisks: already mounted at %s", mount_point)

            self.known_mounts[path] = mount_point
            if self.validMounts.pathIsValidMountPoint(mount_point):
                self.partitionMounted.emit(mount_point, icon_names, ejectable)

        else:
            logging.debug("Udisks: partition has no file system %s", path)

    def retry_mount(self, fs, fstype) -> str:
        # Variant parameter construction Copyright Bernard Baeyens, and is
        # licensed under GNU General Public License Version 2 or higher.
        # https://github.com/berbae/udisksvm
        list_options = ""
        if fstype == "vfat":
            list_options = "flush"
        elif fstype == "ext2":
            list_options = "sync"
        G_VARIANT_TYPE_VARDICT = GLib.VariantType.new("a{sv}")
        param_builder = GLib.VariantBuilder.new(G_VARIANT_TYPE_VARDICT)
        optname = GLib.Variant.new_string("fstype")  # s
        value = GLib.Variant.new_string(fstype)
        vvalue = GLib.Variant.new_variant(value)  # v
        newsv = GLib.Variant.new_dict_entry(optname, vvalue)  # {sv}
        param_builder.add_value(newsv)
        optname = GLib.Variant.new_string("options")
        value = GLib.Variant.new_string(list_options)
        vvalue = GLib.Variant.new_variant(value)
        newsv = GLib.Variant.new_dict_entry(optname, vvalue)
        param_builder.add_value(newsv)
        vparam = param_builder.end()  # a{sv}

        # Try to mount until it does not fail with "Busy"
        timeout = 10
        while timeout >= 0:
            try:
                return fs.call_mount_sync(vparam, None)
            except GLib.GError as e:
                if "UDisks2.Error.DeviceBusy" not in e.message:
                    raise
                logging.debug("Udisks: Device busy.")
                time.sleep(0.3)
                timeout -= 1
        return ""

    def get_icon_names(self, obj: UDisks.Object) -> list[str]:
        # Get icon information, if possible
        icon_names = []
        if have_gio:
            info = self.udisks.get_object_info(obj)
            icon = info.get_icon()
            if isinstance(icon, Gio.ThemedIcon):
                icon_names = icon.get_names()
        return icon_names

    # Next four class member functions from Damon Lynch, not Canonical
    def _device_removed(self, obj: UDisks.Object) -> None:
        # path here refers to the udev / udisks path, not the mount point
        path = obj.get_object_path()
        if path in self.known_mounts:
            mount_point = self.known_mounts[path]
            del self.known_mounts[path]
            self.partitionUnmounted.emit(mount_point)
        else:
            logging.debug(
                "Taking no action on device removal because of unrecognized path: %s",
                path,
            )

    def get_can_eject(self, obj: UDisks.Object) -> bool:
        block = obj.get_block()
        drive = self._get_drive(block)
        if drive is not None:
            return drive.get_property("ejectable")
        return False

    @staticmethod
    def _object_path(device_path: str) -> str:
        """
        Determine object path used by UDisks2 for device path

        :param device_path: system path of the device to check,
        e.g. /dev/sdc1
        """
        return f"/org/freedesktop/UDisks2/block_devices/{os.path.split(device_path)[1]}"

    def get_device_props(self, device_path: str) -> tuple[list[str], bool]:
        """
        Given a device, get the icon names suggested by udev, and
        determine whether the mount is ejectable or not.
        :param device_path: system path of the device to check,
        e.g. /dev/sdc1
        :return: icon names and eject boolean
        """

        object_path = self._object_path(device_path)
        obj = self.udisks.get_object(object_path)
        if obj is None:
            icon_names = []
            can_eject = False
        else:
            icon_names = self.get_icon_names(obj)
            can_eject = self.get_can_eject(obj)
        return icon_names, can_eject

    def add_device(self, device_path: str, mount_point: str) -> None:
        object_path = self._object_path(device_path)
        self.known_mounts[object_path] = mount_point

    @pyqtSlot(str)
    def unmount_volume(self, mount_point: str) -> None:
        G_VARIANT_TYPE_VARDICT = GLib.VariantType.new("a{sv}")
        param_builder = GLib.VariantBuilder.new(G_VARIANT_TYPE_VARDICT)

        # Variant parameter construction Copyright Bernard Baeyens, and is
        # licensed under GNU General Public License Version 2 or higher.
        # https://github.com/berbae/udisksvm

        optname = GLib.Variant.new_string("force")
        value = GLib.Variant.new_boolean(False)
        vvalue = GLib.Variant.new_variant(value)
        newsv = GLib.Variant.new_dict_entry(optname, vvalue)
        param_builder.add_value(newsv)

        vparam = param_builder.end()  # a{sv}

        path = None
        # Get the path from the dict we keep of known mounts
        for key, value in self.known_mounts.items():
            if value == mount_point:
                path = key
                break
        if path is None:
            logging.error(
                "Could not find UDisks2 path used to be able to unmount %s", mount_point
            )

        fs = None
        for obj in self.manager.get_objects():
            opath = obj.get_object_path()
            if path == opath:
                fs = obj.get_filesystem()
        if fs is None:
            logging.error(
                "Could not find UDisks2 filesystem used to be able to unmount %s",
                mount_point,
            )

        logging.debug("Unmounting %s...", mount_point)
        try:
            fs.call_unmount(
                vparam, None, self.umount_volume_callback, (mount_point, fs)
            )
        except GLib.GError:
            value = sys.exc_info()[1]
            logging.error("Unmounting failed with error:")
            logging.error("%s", value)

    def umount_volume_callback(
        self,
        source_object: UDisks.FilesystemProxy,
        result: Gio.AsyncResult,
        user_data: tuple[str, UDisks.Filesystem],
    ) -> None:
        """
        Callback for asynchronous unmount operation.

        :param source_object: the FilesystemProxy object
        :param result: result of the unmount
        :param user_data: mount_point and the file system
        """

        mount_point, fs = user_data

        try:
            if fs.call_unmount_finish(result):
                logging.debug("...successfully unmounted %s", mount_point)
            else:
                # this is the result even when the unmount was unsuccessful
                logging.debug("...possibly failed to unmount %s", mount_point)
        except GLib.GError:
            logging.error("Exception occurred unmounting %s", mount_point)
            logging.exception("Traceback:")
        except Exception:
            logging.error("Exception occurred unmounting %s", mount_point)
            logging.exception("Traceback:")

        self.partitionUnmounted.emit(mount_point)


if have_gio:

    @retry(stop=stop_after_attempt(5), wait=wait_fixed(1))
    def mount_finish(result: Gio.AsyncResult, volume: Gio.Volume) -> bool:
        try:
            return volume.mount_finish(result)
        except gi.repository.GLib.GError as inst:
            logging.warning(
                "Exception occurred unmounting %s: %s",
                volume.get_name() or "",
                str(inst),
            )
            raise
        except Exception as inst:
            logging.error(
                "Unknown exception occurred unmounting %s: %s",
                volume.get_name() or "",
                str(inst),
            )
            raise

    class GVolumeMonitor(QObject):
        r"""
        Monitor the mounting or unmounting of cameras or partitions
        using Gnome's GIO/GVFS. Unmount cameras automatically mounted
        by GVFS.

        Raises a signal if a volume has been inserted, but will not be
        automatically mounted. This is important because this class
        is monitoring mounts, and if the volume is not mounted, it will
        go unnoticed.
        """

        # result (unmount succeeded or not), camera model, port, post unmount action
        cameraUnmounted = pyqtSignal(bool, str, str, PostCameraUnmountAction)

        cameraMounted = pyqtSignal()

        # path, icon names, volume can eject
        partitionMounted = pyqtSignal(str, "PyQt_PyObject", bool)

        # path
        partitionUnmounted = pyqtSignal(str)

        volumeAddedNoAutomount = pyqtSignal()
        cameraPossiblyRemoved = pyqtSignal()

        # device path
        cameraVolumeAdded = pyqtSignal(str)

        def __init__(self, validMounts: ValidMounts, prefs) -> None:
            super().__init__()
            self.prefs = prefs
            self.vm = Gio.VolumeMonitor.get()
            self.vm.connect("mount-added", self.mountAdded)
            self.vm.connect("volume-added", self.volumeAdded)
            self.vm.connect("mount-removed", self.mountRemoved)
            self.vm.connect("volume-removed", self.volumeRemoved)
            self.portSearch = re.compile(r"usb:([\d]+),([\d]+)")
            self.scsiPortSearch = re.compile(r"usbscsi:(.+)")
            self.possibleCamera = re.compile(r"/usb/([\d]+)/([\d]+)")
            self.validMounts = validMounts
            # device_path: volume_name
            self.camera_volumes_added = dict()  # type: dict[str, str]
            self.camera_volumes_mounted = set()  # type: set[str]

            self.manually_mounted_volumes = set()  # type: set[Gio.Volume]

        @staticmethod
        def mountMightBeCamera(mount: Gio.Mount) -> bool:
            """
            :param mount: the mount to check
            :return: True if the mount needs to be checked if it is a camera
            """
            return not mount.is_shadowed() and mount.get_volume() is not None

        def unixDevicePathIsCamera(self, devname: str) -> bool:
            """
            Test if the device at unix device path devname is a camera
            :param devname: Gio.VOLUME_IDENTIFIER_KIND_UNIX_DEVICE device path
             e.g. '/dev/bus/usb/001/003'
            :return: True if camera else False
            """

            return self.possibleCamera.search(devname) is not None and udev_is_camera(
                devname
            )

        def ptpCameraMountPoint(self, model: str, port: str) -> Gio.Mount | None:
            """
            :return: the mount point of the PTP / MTP camera, if it is mounted,
             else None. If camera is not mounted with PTP / MTP, None is
             returned.
            """

            p = self.portSearch.match(port)
            if p is not None:
                p1 = p.group(1)
                p2 = p.group(2)
                device_path = f"/dev/bus/usb/{p1}/{p2}"
                return self.cameraMountPointByUnixDevice(device_path=device_path)
            else:
                p = self.scsiPortSearch.match(port)
                if p is None:
                    logging.error("Unknown camera mount method %s %s", model, port)
                return None

        def cameraMountPointByUnixDevice(self, device_path: str) -> Gio.Mount | None:
            """
            :return: the mount point of the PTP / MTP camera, if it is mounted,
             else None. If camera is not mounted with PTP / MTP, None is
             returned.
            """

            to_unmount = None

            for mount in self.vm.get_mounts():
                if self.mountMightBeCamera(mount):
                    identifier = mount.get_volume().get_identifier(
                        Gio.VOLUME_IDENTIFIER_KIND_UNIX_DEVICE
                    )
                    if device_path == identifier:
                        to_unmount = mount
                        break
            return to_unmount

        @pyqtSlot(str, str, bool, bool, int)
        def reUnmountCamera(
            self,
            model: str,
            port: str,
            post_unmount_action: PostCameraUnmountAction,
            attempt_no: int,
        ) -> None:
            logging.info(
                "Attempt #%s to unmount camera %s on port %s",
                attempt_no + 1,
                model,
                port,
            )
            self.unmountCamera(
                model=model,
                port=port,
                post_unmount_action=post_unmount_action,
                attempt_no=attempt_no,
            )

        def unmountCamera(
            self,
            model: str,
            port: str,
            post_unmount_action: PostCameraUnmountAction,
            mount_point: Gio.Mount | None = None,
            attempt_no: int | None = 0,
        ) -> bool:
            """
            Unmount camera mounted on gvfs mount point, if it is
            mounted. If not mounted, ignore.

            :param model: model as returned by libgphoto2
            :param port: port as returned by libgphoto2, in format like
             usb:001,004
            :param download_starting: if True, the unmount is occurring
             because a download has been initiated.
            :param mount_point: if not None, try umounting from this
             mount point without scanning for it first
            :return: True if an unmount operation has been initiated,
             else returns False.
            """

            if mount_point is None:
                to_unmount = self.ptpCameraMountPoint(model, port)
            else:
                to_unmount = mount_point

            if to_unmount is not None:
                logging.debug("GIO: Attempting to unmount %s...", model)
                to_unmount.unmount_with_operation(
                    0,
                    None,
                    None,
                    self.unmountCameraCallback,
                    (model, port, post_unmount_action, attempt_no),
                )
                return True

            return False

        def unmountCameraCallback(
            self,
            mount: Gio.Mount,
            result: Gio.AsyncResult,
            user_data: tuple[str, str, PostCameraUnmountAction, bool],
        ) -> None:
            """
            Called by the asynchronous unmount operation.
            When complete, emits a signal indicating operation
            success, and the camera model and port
            :param mount: camera mount
            :param result: result of the unmount process
            :param user_data: model and port of the camera being
            unmounted, in the format of libgphoto2
            """

            model, port, post_unmount_action, attempt_no = user_data
            try:
                if mount.unmount_with_operation_finish(result):
                    logging.debug(f"...successfully unmounted {model}")
                    self.cameraUnmounted.emit(True, model, port, post_unmount_action)
                else:
                    logging.debug(f"...failed to unmount {model}")
                    self.cameraUnmounted.emit(False, model, port, post_unmount_action)
            except GLib.GError as e:
                if e.code == 26 and attempt_no < 10:
                    attempt_no += 1
                    QTimer.singleShot(
                        1000,
                        lambda: self.reUnmountCamera(
                            model, port, post_unmount_action, attempt_no
                        ),
                    )
                else:
                    logging.error(f"Exception occurred unmounting {model}")
                    logging.exception("Traceback:")
                    self.cameraUnmounted.emit(False, model, port, post_unmount_action)

        def unmountVolume(self, path: str) -> None:
            """
            Unmounts the volume represented by the path. If no volume is found
            representing that path, nothing happens.

            :param path: path of the volume. It should not end with os.sep.
            """

            for mount in self.vm.get_mounts():
                root = mount.get_root()
                if root is not None:
                    mpath = root.get_path()
                    if path == mpath:
                        logging.info("Attempting to unmount %s...", path)
                        mount.unmount_with_operation(
                            0, None, None, self.unmountVolumeCallback, path
                        )
                        break

        @staticmethod
        def unmountVolumeCallback(
            mount: Gio.Mount, result: Gio.AsyncResult, user_data: str
        ) -> None:
            """
            Called by the asynchronous unmount operation.

            :param mount: volume mount
            :param result: result of the unmount process
            :param user_data: the path of the device unmounted
            """
            path = user_data

            try:
                if mount.unmount_with_operation_finish(result):
                    logging.info("...successfully unmounted volume %s", path)
                else:
                    logging.info("...failed to unmount volume %s", path)
            except GLib.GError as e:
                if e.code == 16:
                    logging.debug("...backend currently unmounting volume %s...", path)
                elif e.code == 26:
                    logging.debug(
                        "...did not yet unmount volume %s because it is busy..."
                    )
                    # TODO investigate if should try again to unmount the volume,
                    #  similar to unmountCameraCallback()
                else:
                    logging.error("Exception occurred unmounting volume %s", path)
                    logging.exception("Traceback:")

        @staticmethod
        def mountIsAppleFileConduit(mount: Gio.Mount, path: str) -> bool:
            if path:
                logging.debug(
                    f"GIO: Looking for Apple File Conduit (AFC) at mount {path}"
                )
                path, folder_name = os.path.split(path)
                if folder_name and folder_name.startswith("afc:host="):
                    return True
            return False

        def mountVolume(self, volume: Gio.Volume) -> None:
            logging.debug("Attempting to mount %s", volume.get_name())
            self.manually_mounted_volumes.add(volume)
            volume.mount(0, None, None, self.mountVolumeCallback, volume)

        def mountVolumeCallback(
            self, source_object, result: Gio.AsyncResult, volume: Gio.Volume
        ) -> None:
            self.manually_mounted_volumes.remove(volume)

            if mount_finish(result, volume):
                logging.debug("%s was successfully manually mounted", volume.get_name())
                self.mountAdded(self.vm, volume.get_mount())
            else:
                logging.debug("%s failed to mount", volume.get_name())

        def mountIsCamera(self, mount: Gio.Mount, path: str | None = None) -> bool:
            """
            Determine if the mount refers to a camera by checking the
            path to see if gphoto2 or mtp is in the last folder in the
            root path.

            Does not query udev, deliberately. This can be called when device
            is being unmounted. Unclear if the device is still on the system
            at this point, or how realible that is even if it is.

            :param mount: mount to check
            :param path: optional mount path if already determined
            :return: True if mount refers to a camera, else False
            """

            if self.mountMightBeCamera(mount):
                if path is None:
                    root = mount.get_root()
                    if root is None:
                        logging.warning(
                            "Unable to get mount root for %s", mount.get_name()
                        )
                    else:
                        path = root.get_path()
                if path:
                    logging.debug(f"GIO: Looking for camera at mount {path}")
                    # check last two levels of the path name, as it might be in a
                    # format like
                    # /run/..../gvfs/gphoto2:host=Canon_Inc._Canon_Digital_Camera/store_00010001 # noqa:E501
                    for i in (1, 2):
                        path, folder_name = os.path.split(path)
                        if folder_name:
                            for s in ("gphoto2:host=", "mtp:host="):
                                if folder_name.startswith(s):
                                    return True
            return False

        def mountIsPartition(self, mount: Gio.Mount, path: str | None = None) -> bool:
            """
            Determine if the mount point is that of a valid partition,
            i.e. is mounted in a valid location, which is under one of
            self.validMountDirs
            :param mount: the mount to examine
            :param path: optional mount path if already determined
            :return: True if the mount is a valid partiion
            """

            if path is None:
                root = mount.get_root()
                if root is None:
                    logging.warning("Unable to get mount root for %s", mount.get_name())
                else:
                    path = root.get_path()
            if path:
                logging.debug(f"GIO: Looking for valid partition at mount {path}")
                if self.validMounts.pathIsValidMountPoint(path):
                    logging.debug(f"GIO: partition found at {path}")
                    return True
            if path is not None:
                logging.debug(f"GIO: partition is not valid mount: {path}")
            return False

        def mountAdded(self, volumeMonitor, mount: Gio.Mount) -> None:
            """
            Determine if mount is valid partition or is a camera, or something
            else.

            :param volumeMonitor: not used
            :param mount: the mount to examine
            """

            if mount.get_volume() in self.manually_mounted_volumes:
                logging.debug(
                    "Waiting for manual mount of %s to complete",
                    mount.get_volume().get_name(),
                )
                return

            logging.debug("Examining mount %s", mount.get_name())
            try:
                identifier = mount.get_volume().get_identifier(
                    Gio.VOLUME_IDENTIFIER_KIND_UNIX_DEVICE
                )
                if identifier in self.camera_volumes_added:
                    logging.debug(
                        "%s is now mounted", self.camera_volumes_added[identifier]
                    )
                    self.camera_volumes_mounted.add(identifier)
                    self.cameraMounted.emit()
                    return
            except Exception:
                pass

            try:
                path = mount.get_root().get_path()
            except Exception:
                logging.warning("Unable to get mount path for %s", mount.get_name())
            else:
                if self.mountIsAppleFileConduit(mount, path):
                    # An example of an AFC volume is the "Documents" mount for an
                    # iPhone, which in contrast to the gphoto2 mount for the same device
                    logging.debug("Apple File Conduit (AFC) mount detected at %s", path)
                    logging.info("Attempting to unmount %s...", path)
                    mount.unmount_with_operation(
                        0, None, None, self.unmountVolumeCallback, path
                    )
                elif self.mountIsCamera(mount, path):
                    # Can be called on startup if the camera was already mounted in GIO
                    # before the program started. In that case, previous check would not
                    # have detected the camera.
                    self.cameraMounted.emit()
                elif self.mountIsPartition(mount, path):
                    icon_names = self.getIconNames(mount)
                    self.partitionMounted.emit(
                        mount.get_root().get_path(), icon_names, mount.can_eject()
                    )

        def mountRemoved(self, volumeMonitor, mount: Gio.Mount) -> None:
            if not self.mountIsCamera(mount) and self.mountIsPartition(mount):
                logging.debug("GIO: %s has been unmounted", mount.get_name())
                self.partitionUnmounted.emit(mount.get_root().get_path())

        def volumeAdded(self, volumeMonitor, volume: Gio.Volume) -> None:
            volume_name = volume.get_name()
            should_automount = volume.should_automount()
            logging.debug(
                "GIO: Volume added %s. Automount: %s",
                volume_name,
                should_automount,
            )

            if not should_automount:
                logging.debug(
                    "%s has probably been removed: do not automount", volume_name
                )
                return

            if not self.prefs.auto_mount:
                logging.debug(
                    "Not checking mount status for %s because auto mount preference "
                    "is off",
                    volume_name,
                )
                return

            # Even if volume.should_automount(), the volume in fact may not be mounted
            # automatically. It's a bug that has shown up at least twice!

            device_path = volume.get_identifier(Gio.VOLUME_IDENTIFIER_KIND_UNIX_DEVICE)
            if device_path is None:
                logging.debug("%s is not a Unix Device", volume_name)
            else:
                try:
                    is_camera = self.unixDevicePathIsCamera(device_path)
                except TypeError:
                    logging.debug(
                        "Unexpected device path for %s. Type %s",
                        volume_name,
                        type(device_path),
                    )
                else:
                    if is_camera:
                        self.camera_volumes_added[device_path] = volume_name
                        logging.debug("%s is a camera at %s", volume_name, device_path)
                        # Time is in milliseconds; 3000 is 3 seconds.
                        QTimer.singleShot(
                            3000, lambda: self.cameraVolumeAddedCheckMount(device_path)
                        )
                    else:
                        uuid = volume.get_uuid()
                        logging.debug(
                            "%s is a device at %s with UUID %s",
                            volume_name,
                            device_path,
                            uuid,
                        )
                        QTimer.singleShot(
                            3000, lambda: self.deviceVolumeAddedCheckMount(volume)
                        )

        def cameraVolumeAddedCheckMount(self, device_path) -> None:
            if device_path not in self.camera_volumes_mounted:
                logging.debug(
                    "%s had not been automatically mounted. Will initiate camera scan.",
                    self.camera_volumes_added[device_path],
                )
                self.cameraVolumeAdded.emit(device_path)
            else:
                logging.debug(
                    "%s had been automatically mounted",
                    self.camera_volumes_added[device_path],
                )

        def deviceVolumeAddedCheckMount(self, volume: Gio.Volume) -> None:
            mount = volume.get_mount()
            # Double check that it's in the list of mounts
            mounted = mount in self.vm.get_mounts() if mount is not None else False

            if not mounted:
                logging.debug(
                    "%s has not been automatically mounted. Will initiate mount.",
                    volume.get_name(),
                )
                self.mountVolume(volume)
            else:
                logging.debug("%s was automatically mounted", volume.get_name())

        def volumeRemoved(self, volumeMonitor, volume: Gio.Volume) -> None:
            logging.debug("GIO: %s volume removed", volume.get_name())
            if volume.get_activation_root() is not None:
                logging.debug("GIO: %s might be a camera", volume.get_name())
                self.cameraPossiblyRemoved.emit()

        @staticmethod
        def getIconNames(mount: Gio.Mount) -> list[str]:
            """
            Get icons for the mount from theme

            :param mount:
            :return:
            """
            icon_names = []
            icon = mount.get_icon()
            if isinstance(icon, Gio.ThemedIcon):
                icon_names = icon.get_names()

            return icon_names

        def getProps(self, path: str) -> tuple[list[str] | None, bool | None]:
            """
            Given a mount's path, get the icon names suggested by the
            volume monitor, and determine whether the mount is
            ejectable or not.
            :param path: the path of mount to check
            :return: icon names and eject boolean
            """

            for mount in self.vm.get_mounts():
                root = mount.get_root()
                if root is not None:
                    p = root.get_path()
                    if path == p:
                        icon_names = self.getIconNames(mount)
                        return (icon_names, mount.can_eject())
            return (None, None)


def _get_info_size_value(info: Gio.FileInfo, attr: str) -> int:
    if info.get_attribute_data(attr).type == Gio.FileAttributeType.UINT64:
        return info.get_attribute_uint64(attr)
    else:
        return info.get_attribute_uint32(attr)


def get_mount_size(mount: QStorageInfo) -> tuple[int, int]:
    """
    Uses GIO to get bytes total and bytes free (available) for the mount that a
    path is in.

    :param path: path located anywhere in the mount
    :return: bytes_total, bytes_free
    """

    bytes_free = mount.bytesAvailable()
    bytes_total = mount.bytesTotal()

    if bytes_total or not have_gio:
        return bytes_total, bytes_free

    path = mount.rootPath()

    logging.debug("Using GIO to query file system attributes for %s...", path)
    p = Gio.File.new_for_path(os.path.abspath(path))
    info = p.query_filesystem_info(
        ",".join(
            (Gio.FILE_ATTRIBUTE_FILESYSTEM_SIZE, Gio.FILE_ATTRIBUTE_FILESYSTEM_FREE)
        )
    )
    logging.debug("...query of file system attributes for %s completed", path)
    bytes_total = _get_info_size_value(info, Gio.FILE_ATTRIBUTE_FILESYSTEM_SIZE)
    bytes_free = _get_info_size_value(info, Gio.FILE_ATTRIBUTE_FILESYSTEM_FREE)
    return bytes_total, bytes_free



# This file is part of ReText
# Copyright: 2015-2017 Dmitry Shachnev
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This is implementation of XSettings specification, described at
# <http://standards.freedesktop.org/xsettings-spec/xsettings-spec-0.5.html>

# Reformatted in 2022 by Damon Lynch, using black

import ctypes
import ctypes.util
import struct


class _xcb_reply_t(ctypes.Structure):
    # this can be used instead of xcb_intern_atom_reply_t,
    # xcb_get_selection_owner_reply_t, etc
    _fields_ = [
        ("response_type", ctypes.c_uint8),
        ("pad0", ctypes.c_uint8),
        ("sequence", ctypes.c_uint16),
        ("length", ctypes.c_uint32),
        ("payload", ctypes.c_uint32),
    ]


class _xcb_cookie_t(ctypes.Structure):
    # this can be used instead of xcb_intern_atom_cookie_t,
    # xcb_get_selection_owner_cookie_t, etc
    _fields_ = [("sequence", ctypes.c_uint)]


_xcb_error_messages = [
    None,
    "XCB error: socket, pipe and other stream error",
    "XCB connection closed: extension unsupported",
    "XCB connection closed: insufficient memory",
    "XCB connection closed: request length exceeded",
    "XCB connection closed: DISPLAY parse error",
    "XCB connection closed: invalid screen",
]


class XSettingsError(RuntimeError):
    pass


class XSettingsParseError(XSettingsError):
    pass


def get_raw_xsettings(display=0):
    # initialize the libraries
    xcb_library_name = ctypes.util.find_library("xcb")
    if xcb_library_name is None:
        raise XSettingsError("Xcb library not found")
    xcb = ctypes.CDLL(xcb_library_name)

    c_library_name = ctypes.util.find_library("c")
    if c_library_name is None:
        raise XSettingsError("C library not found")
    c = ctypes.CDLL(c_library_name)

    # set some args and return types
    c.free.argtypes = [ctypes.c_void_p]
    c.free.restype = None
    xcb.xcb_connect.argtypes = [ctypes.c_char_p, ctypes.POINTER(ctypes.c_int)]
    xcb.xcb_connect.restype = ctypes.c_void_p
    xcb.xcb_connection_has_error.argtypes = [ctypes.c_void_p]
    xcb.xcb_connection_has_error.restype = ctypes.c_int
    xcb.xcb_disconnect.argtypes = [ctypes.c_void_p]
    xcb.xcb_disconnect.restype = None
    xcb.xcb_intern_atom.argtypes = [
        ctypes.c_void_p,
        ctypes.c_uint8,
        ctypes.c_uint16,
        ctypes.c_char_p,
    ]
    xcb.xcb_intern_atom.restype = _xcb_cookie_t
    xcb.xcb_intern_atom_reply.argtypes = [
        ctypes.c_void_p,
        _xcb_cookie_t,
        ctypes.c_void_p,
    ]
    xcb.xcb_intern_atom_reply.restype = ctypes.POINTER(_xcb_reply_t)
    xcb.xcb_get_selection_owner.argtypes = [ctypes.c_void_p, ctypes.c_uint32]
    xcb.xcb_get_selection_owner.restype = _xcb_cookie_t
    xcb.xcb_get_selection_owner_reply.argtypes = [
        ctypes.c_void_p,
        _xcb_cookie_t,
        ctypes.c_void_p,
    ]
    xcb.xcb_get_selection_owner_reply.restype = ctypes.POINTER(_xcb_reply_t)
    xcb.xcb_get_property.argtypes = [
        ctypes.c_void_p,
        ctypes.c_uint8,
        ctypes.c_uint32,
        ctypes.c_uint32,
        ctypes.c_uint32,
        ctypes.c_uint32,
        ctypes.c_uint32,
    ]
    xcb.xcb_get_property.restype = _xcb_cookie_t
    xcb.xcb_get_property_reply.argtypes = [
        ctypes.c_void_p,
        _xcb_cookie_t,
        ctypes.c_void_p,
    ]
    xcb.xcb_get_property_reply.restype = ctypes.c_void_p
    xcb.xcb_get_property_value_length.argtypes = [ctypes.c_void_p]
    xcb.xcb_get_property_value_length.restype = ctypes.c_int
    xcb.xcb_get_property_value.argtypes = [ctypes.c_void_p]
    xcb.xcb_get_property_value.restype = ctypes.c_void_p

    # open the connection
    connection = xcb.xcb_connect(None, None)
    error = xcb.xcb_connection_has_error(connection)
    if error:
        raise XSettingsError(_xcb_error_messages[error])

    # get selection atom cookie
    buffer = ("_XSETTINGS_S%d" % display).encode()
    cookie = xcb.xcb_intern_atom(connection, 0, len(buffer), buffer)

    # get selection atom reply
    reply = xcb.xcb_intern_atom_reply(connection, cookie, None)
    selection_atom = reply.contents.payload
    c.free(reply)

    # get selection owner cookie
    cookie = xcb.xcb_get_selection_owner(connection, selection_atom)

    # get selection owner reply
    reply = xcb.xcb_get_selection_owner_reply(connection, cookie, None)
    window = reply.contents.payload
    c.free(reply)

    # get settings atom cookie
    buffer = b"_XSETTINGS_SETTINGS"
    cookie = xcb.xcb_intern_atom(connection, 0, len(buffer), buffer)

    # get settings atom reply
    reply = xcb.xcb_intern_atom_reply(connection, cookie, None)
    settings_atom = reply.contents.payload
    c.free(reply)

    # get property cookie
    cookie = xcb.xcb_get_property(connection, 0, window, settings_atom, 0, 0, 0x2000)

    # get property reply
    reply = xcb.xcb_get_property_reply(connection, cookie, None)
    if reply is not None:
        length = xcb.xcb_get_property_value_length(reply)
        pointer = xcb.xcb_get_property_value(reply) if length else None
        result = ctypes.string_at(pointer, length)
        c.free(reply)

    # close the connection
    xcb.xcb_disconnect(connection)

    # handle possible errors
    if reply is None or not length:
        raise XSettingsError("XSettings not available")

    return result


def parse_xsettings(raw_xsettings):
    if len(raw_xsettings) < 12:
        raise XSettingsParseError("length < 12")

    if raw_xsettings[0] not in (0, 1):
        raise XSettingsParseError("wrong order byte: %d" % raw_xsettings[0])
    byte_order = "<>"[raw_xsettings[0]]
    settings_count = struct.unpack(byte_order + "I", raw_xsettings[8:12])[0]

    TypeInteger, TypeString, TypeColor = range(3)
    result = {}

    raw_xsettings = raw_xsettings[12:]
    offset = 0
    for i in range(settings_count):
        setting_type = raw_xsettings[offset]
        offset += 2
        name_length = struct.unpack(
            byte_order + "H", raw_xsettings[offset : offset + 2]
        )[0]
        offset += 2
        name = raw_xsettings[offset : offset + name_length]
        offset += name_length
        if offset & 3:
            offset += 4 - (offset & 3)
        offset += 4  # skip last-change-serial

        if setting_type == TypeInteger:
            value = struct.unpack(byte_order + "I", raw_xsettings[offset : offset + 4])[
                0
            ]
            offset += 4
        elif setting_type == TypeString:
            value_length = struct.unpack(
                byte_order + "I", raw_xsettings[offset : offset + 4]
            )[0]
            offset += 4
            value = raw_xsettings[offset : offset + value_length]
            offset += value_length
            if offset & 3:
                offset += 4 - (offset & 3)
        elif setting_type == TypeColor:
            value = struct.unpack(
                byte_order + "HHHH", raw_xsettings[offset : offset + 8]
            )
            offset += 8
        else:
            raise XSettingsParseError("Wrong setting type: %d" % setting_type)
        result[name] = value
    return result


def get_xsettings(display=0):
    raw_xsettings = get_raw_xsettings(display)
    return parse_xsettings(raw_xsettings)



# Copyright (C) 2020-2021 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader. If not,
# see <http://www.gnu.org/licenses/>.

"""
Handle when the user clicks on a URL in Qt and the URL scheme is file://

The point is to open a file manager which selects the file in the URI, rather than
opening the file directly.
"""


__author__ = "Damon Lynch"
__copyright__ = "Copyright 2020-2021, Damon Lynch"

from PyQt5.QtCore import QObject, QUrl, pyqtSlot
from showinfm import show_in_file_manager


class FileSystemUrlHandler(QObject):
    @pyqtSlot(QUrl)
    def openFileBrowser(self, url: QUrl):
        show_in_file_manager(url.url(options=QUrl.FullyEncoded))



# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Widget containing Header with Toggle Switch, and contains widget that appears or
disappears depending on the toggle switch's state.

Portions modeled on Canonical's QExpander, which is an 'Expander widget
similar to the GtkExpander', Copyright 2012 Canonical Ltd
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"

from PyQt5.QtCore import QSize, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QLayout, QWidget  # noqa: F401

from raphodo.constants import DarkModeHeaderBackgroundName, HeaderBackgroundName
from raphodo.ui.panelview import QPanelView
from raphodo.ui.toggleswitch import QToggleSwitch
from raphodo.ui.viewutils import BlankWidget, is_dark_mode


class QToggleView(QPanelView):
    """
    A header bar with toggle switch over a widget that is switched on/off.
    """

    valueChanged = pyqtSignal(bool)

    def __init__(
        self,
        label: str,
        display_alternate: bool,
        object_name: str,
        toggleToolTip: str | None,
        headerColor: QColor | None = None,
        headerFontColor: QColor | None = None,
        on: bool = True,
        parent: QWidget = None,
    ) -> None:
        if headerColor is None:
            if is_dark_mode():
                headerColor = QColor(DarkModeHeaderBackgroundName)
            else:
                headerColor = QColor(HeaderBackgroundName)
        super().__init__(
            label=label,
            headerColor=headerColor,
            headerFontColor=headerFontColor,
            parent=parent,
        )
        # Override base class definition:
        self.headerLayout.setContentsMargins(5, 0, 5, 0)
        self.setObjectName(object_name)

        if display_alternate:
            self.alternateWidget = BlankWidget()
            layout = self.layout()  # type: QLayout
            layout.addWidget(self.alternateWidget)
        else:
            self.alternateWidget = None

        self.toggleSwitch = QToggleSwitch(background=headerColor, parent=self)
        self.toggleSwitch.valueChanged.connect(self.toggled)
        if toggleToolTip:
            self.toggleSwitch.setToolTip(toggleToolTip)
        self.addHeaderWidget(self.toggleSwitch)
        self.toggleSwitch.setOn(on)

    def addWidget(self, widget: QWidget) -> None:
        super().addWidget(widget)
        self.toggled(0)

    def on(self) -> bool:
        """Return if widget is expanded."""

        return self.toggleSwitch.on()

    def setOn(self, isOn: bool) -> None:
        """Expand the widget or not."""

        self.toggleSwitch.setOn(isOn)

    @pyqtSlot(int)
    def toggled(self, value: int) -> None:
        if self.content is not None:
            self.content.setVisible(self.on())
            if self.alternateWidget is not None:
                self.alternateWidget.setVisible(not self.on())

        self.valueChanged.emit(self.on())

    def minimumSize(self) -> QSize:
        size = super().minimumSize()
        width = size.width()
        height = self.minimumHeight()
        return QSize(width, height)

    def minimumHeight(self) -> int:
        if not self.toggleSwitch.on():
            return self.header.height()
        else:
            # critically important to call updateGeometry(), as minimum height is
            # recalculated *after* sizeHint has been called by the Qt layout manager
            self.updateGeometry()
            return super().minimumSize().height()

    def sizeHint(self) -> QSize:
        return self.minimumSize()
